from datetime import date, datetime, timedelta

from cumulusci.tasks.bulkdata.dates import (
    adjust_relative_dates,
    datetime_from_salesforce,
    salesforce_from_datetime,
)
from cumulusci.tasks.bulkdata.mapping_parser import MappingStep
from cumulusci.tasks.bulkdata.step import DataOperationType


class TestRelativeDates:
    def test_relative_dates(self):
        mapping = MappingStep(
            sf_object="Account", fields=["Some_Date__c"], anchor_date="2020-07-01"
        )

        target = date.today() + timedelta(days=7)
        assert adjust_relative_dates(
            mapping, ([0], [], date.today()), ["2020-07-08"], DataOperationType.INSERT
        ) == [target.isoformat()]

        assert adjust_relative_dates(
            mapping, ([0], [], date.today()), ["2020-07-01"], DataOperationType.INSERT
        ) == [date.today().isoformat()]

        assert adjust_relative_dates(
            mapping, ([0], [], date.today()), [""], DataOperationType.INSERT
        ) == [""]

    def test_relative_dates__extract(self):
        mapping = MappingStep(
            sf_object="Account", fields=["Some_Date__c"], anchor_date="2020-07-01"
        )

        target = mapping.anchor_date + timedelta(days=7)
        input_date = (date.today() + timedelta(days=7)).isoformat()
        assert (
            adjust_relative_dates(
                mapping,
                ([1], [], date.today()),
                ["001000000000000", input_date],
                DataOperationType.QUERY,
            )
            == ["001000000000000", target.isoformat()]
        )

        assert (
            adjust_relative_dates(
                mapping,
                ([1], [], date.today()),
                ["001000000000000", date.today().isoformat()],
                DataOperationType.QUERY,
            )
            == ["001000000000000", mapping.anchor_date.isoformat()]
        )

        assert (
            adjust_relative_dates(
                mapping,
                ([1], [], date.today()),
                ["001000000000000", ""],
                DataOperationType.QUERY,
            )
            == ["001000000000000", ""]
        )

    def test_relative_datetimes(self):
        mapping = MappingStep(
            sf_object="Account", fields=["Some_Datetime__c"], anchor_date="2020-07-01"
        )

        input_dt = datetime_from_salesforce("2020-07-08T09:37:57.373+0000")
        target = datetime.combine(date.today() + timedelta(days=7), input_dt.time())
        assert (
            adjust_relative_dates(
                mapping,
                ([], [0], date.today()),
                [salesforce_from_datetime(input_dt)],
                DataOperationType.INSERT,
            )
            == [salesforce_from_datetime(target)]
        )

        now = datetime.combine(mapping.anchor_date, datetime.now().time())
        assert (
            adjust_relative_dates(
                mapping,
                ([], [0], date.today()),
                [salesforce_from_datetime(now)],
                DataOperationType.INSERT,
            )
            == [salesforce_from_datetime(datetime.combine(date.today(), now.time()))]
        )

        assert adjust_relative_dates(
            mapping, ([], [0], date.today()), [""], DataOperationType.INSERT
        ) == [""]

    def test_relative_datetimes_extract(self):
        mapping = MappingStep(
            sf_object="Account", fields=["Some_Datetime__c"], anchor_date="2020-07-01"
        )

        input_dt = datetime.now() + timedelta(days=7)
        target = datetime.combine(
            mapping.anchor_date + timedelta(days=7), input_dt.time()
        )
        assert (
            adjust_relative_dates(
                mapping,
                ([], [1], date.today()),
                ["001000000000000", salesforce_from_datetime(input_dt)],
                DataOperationType.QUERY,
            )
            == ["001000000000000", salesforce_from_datetime(target)]
        )



import pytest
import responses

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.tasks.salesforce.profiles import CreateBlankProfile

from .util import create_task

RESPONSE_FAULT = """<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:sf="urn:fault.partner.soap.sforce.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<soapenv:Body>
    <soapenv:Fault>
        <faultcode>sf:INVALID_SESSION_ID</faultcode>
        <faultstring>INVALID_SESSION_ID: Invalid Session ID found in SessionHeader: Illegal Session</faultstring>
        <detail>
        <sf:UnexpectedErrorFault xsi:type="sf:UnexpectedErrorFault">
            <sf:exceptionCode>INVALID_SESSION_ID</sf:exceptionCode>
            <sf:exceptionMessage>Invalid Session ID found in SessionHeader: Illegal Session</sf:exceptionMessage>
        </sf:UnexpectedErrorFault>
        </detail>
    </soapenv:Fault>
</soapenv:Body>
</soapenv:Envelope>"""

RESPONSE_ERROR = """<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns="urn:partner.soap.sforce.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<soapenv:Body>
    <createResponse>
        <result>
        <errors>
            <message>insufficient access rights on cross-reference id</message>
            <statusCode>INSUFFICIENT_ACCESS_ON_CROSS_REFERENCE_ENTITY</statusCode>
        </errors>
        <id xsi:nil="true"/>
        <success>false</success>
        </result>
    </createResponse>
</soapenv:Body>
</soapenv:Envelope>"""

RESPONSE_ERROR_FIELD = """<?xml version="1.0" encoding="UTF-8"?>
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns="urn:partner.soap.sforce.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
<soapenv:Header>
    <LimitInfoHeader>
        <limitInfo>
        <current>43</current>
        <limit>5000000</limit>
        <type>API REQUESTS</type>
        </limitInfo>
    </LimitInfoHeader>
</soapenv:Header>
<soapenv:Body>
    <createResponse>
        <result>
        <errors>
            <fields>Name</fields>
            <message>The profile name is already in use: Name</message>
            <statusCode>FIELD_INTEGRITY_EXCEPTION</statusCode>
        </errors>
        <id xsi:nil="true" />
        <success>false</success>
        </result>
    </createResponse>
</soapenv:Body>
</soapenv:Envelope>"""

RESPONSE_SUCCESS = """<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns="urn:partner.soap.sforce.com">
<soapenv:Body>
    <createResponse>
        <result>
        <id>001R0000029IyDPIA0</id>
        <success>true</success>
        </result>
    </createResponse>
</soapenv:Body>
</soapenv:Envelope>"""


@responses.activate
def test_run_task_success():
    query_url = "https://test.salesforce.com/services/data/v53.0/query/"

    task = create_task(
        CreateBlankProfile,
        {
            "license": "Foo",
            "name": "Test Profile Name",
            "description": "Have fun stormin da castle",
        },
    )
    task.org_config._latest_api_version = "53.0"

    responses.add(
        responses.GET,
        query_url,
        json={
            "done": True,
            "totalSize": 1,
            "records": [
                {
                    "attributes": {
                        "type": "UserLicense",
                        "url": "/services/data/v53.0/sobjects/UserLicense/10056000000VGjUAAW",
                    },
                    "Id": "10056000000VGjUAAW",
                    "Name": "Salesforce",
                }
            ],
        },
    )
    responses.add(
        responses.POST,
        "https://test.salesforce.com/services/Soap/u/53.0/ORG_ID",
        RESPONSE_SUCCESS,
    )
    result = task._run_task()
    assert "001R0000029IyDPIA0" == result
    assert responses.calls[0].request.params == {
        "q": "SELECT Id, Name FROM UserLicense WHERE Name = 'Foo' LIMIT 1"
    }
    soap_body = responses.calls[1].request.body
    assert "<Name>Test Profile Name</Name>" in str(soap_body)
    assert "<UserLicenseId>10056000000VGjUAAW</UserLicenseId>" in str(soap_body)
    assert "<Description>Have fun stormin da castle</Description>" in str(soap_body)


@responses.activate
def test_run_task_fault():
    task = create_task(
        CreateBlankProfile,
        {
            "license_id": "10056000000VGjUAAW",
            "name": "Test Profile",
            "description": "This is the description",
        },
    )
    task.org_config._latest_api_version = "53.0"

    responses.add(
        responses.POST,
        "https://test.salesforce.com/services/Soap/u/53.0/ORG_ID",
        RESPONSE_FAULT,
    )
    with pytest.raises(MetadataApiError) as e:
        task._run_task()

    assert "Invalid Session ID found" in str(e)


@responses.activate
def test_run_task_field_error():
    task = create_task(
        CreateBlankProfile,
        {
            "license_id": "10056000000VGjUAAW",
            "name": "Test Profile",
            "description": "This is the description",
        },
    )
    task.org_config._latest_api_version = "53.0"
    responses.add(
        responses.POST,
        "https://test.salesforce.com/services/Soap/u/53.0/ORG_ID",
        RESPONSE_ERROR_FIELD,
    )

    with pytest.raises(MetadataApiError) as e:
        task._run_task()
    assert "The profile name is already in use: Name" in str(e)


@responses.activate
def test_run_task_error():
    task = create_task(
        CreateBlankProfile,
        {
            "license_id": "10056000000VGjUAAW",
            "name": "Test Profile",
            "description": "This is the description",
        },
    )
    task.org_config._latest_api_version = "53.0"

    responses.add(
        responses.POST,
        "https://test.salesforce.com/services/Soap/u/53.0/ORG_ID",
        RESPONSE_ERROR,
    )

    with pytest.raises(MetadataApiError) as e:
        task._run_task()
    assert "insufficient access rights on cross-reference id" in str(e)


def test_task_options_error():
    with pytest.raises(TaskOptionsError):
        create_task(
            CreateBlankProfile,
            {
                "name": "Foo",
                "description": "Foo",
            },
        )



import unittest
from unittest.mock import Mock, call

from cumulusci.core.exceptions import CumulusCIException, SalesforceException
from cumulusci.tasks.salesforce.network_member_group import CreateNetworkMemberGroups
from cumulusci.tasks.salesforce.tests.util import create_task


class TestCreateNetworkMemberGroups(unittest.TestCase):
    """
    Unit tests cumulusci.tasks.salesforce.network_member_group.CreateNetworkMemberGroups.
    """

    def test_get_network_id__no_network_found(self):
        network_name = "network_name"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task.sf = Mock()
        task.sf.query_all.Mock()
        task.sf.query_all.return_value = {"records": []}
        task.format_soql = Mock()

        # Execute the test.
        with self.assertRaises(SalesforceException) as context:
            task._get_network_id(network_name)

        # Assert scenario execute as expected.
        self.assertEqual(
            f'No Network record found with Name "{network_name}"',
            context.exception.args[0],
        )

        task.sf.query_all.assert_called_once_with(
            f"SELECT Id FROM Network WHERE Name = '{network_name}' LIMIT 1"
        )

    def test_get_network_id__network_found(self):
        network_name = "network_name"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task.sf = Mock()
        task.sf.query_all.return_value = {"records": [{"Id": "NetworkId"}]}
        task.format_soql = Mock()

        expected = task.sf.query_all.return_value["records"][0]["Id"]

        # Execute the test.
        actual = task._get_network_id(network_name)

        # Assert scenario execute as expected.
        self.assertEqual(expected, actual)

        task.sf.query_all.assert_called_once_with(
            f"SELECT Id FROM Network WHERE Name = '{network_name}' LIMIT 1"
        )

    def test_get_network_member_group_parent_ids(self):
        network_name = "network_name"
        network_id = "network_id"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task.sf = Mock()
        task.sf.query_all.return_value = {
            "records": [
                {"ParentId": "0"},
                {"ParentId": "2"},
                {"ParentId": "3"},
                {"ParentId": "1"},
            ]
        }

        expected = set(["0", "1", "2", "3"])

        # Execute the test.
        actual = task._get_network_member_group_parent_ids(network_id)

        # Assert scenario execute as expected.
        self.assertEqual(expected, actual)

        task.sf.query_all.assert_called_once_with(
            f"SELECT ParentId FROM NetworkMemberGroup WHERE NetworkId = '{network_id}'"
        )

    def test_get_parent_ids_by_name(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        record_names = [
            "Name_0",
            "Name_1",
            "Name_2",
        ]

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task.sf = Mock()
        task.sf.query_all = Mock(
            return_value={
                "records": [
                    {"Name": "Name_0", "Id": "Id_0"},
                    {"Name": "Name_2", "Id": "Id_2"},
                ]
            }
        )

        expected = {
            "Name_0": "Id_0",
            "Name_1": None,
            "Name_2": "Id_2",
        }

        # Execute the test.
        actual = task._get_parent_ids_by_name(sobject_type, record_names)

        # Assert scenario execute as expected.
        self.assertEqual(expected, actual)

        task.sf.query_all.assert_called_once_with(
            "SELECT Id, Name FROM {} WHERE Name IN ('{}')".format(
                sobject_type,
                "','".join(record_names),
            )
        )

    def test_get_parent_ids_by_label(self):
        network_name = "network_name"
        sobject_type = "PermissionSet"
        record_names = [
            "Name_0",
            "Name_1",
            "Name_2",
        ]

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task.sf = Mock()
        task.sf.query_all.return_value = {
            "records": [
                {"Label": "Name_0", "Id": "Id_0"},
                {"Label": "Name_2", "Id": "Id_2"},
            ]
        }

        expected = {
            "Name_0": "Id_0",
            "Name_1": None,
            "Name_2": "Id_2",
        }

        # Execute the test.
        actual = task._get_parent_ids_by_name(sobject_type, record_names)

        # Assert scenario execute as expected.
        self.assertEqual(expected, actual)

        task.sf.query_all.assert_called_once_with(
            "SELECT Id, Label FROM {} WHERE Label IN ('{}')".format(
                sobject_type,
                "','".join(record_names),
            )
        )

    def test_process_parent__no_names(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        for record_names in [
            None,
            [],
        ]:

            task = create_task(
                CreateNetworkMemberGroups, {"network_name": network_name}
            )

            parent_ids_by_name = Mock()

            task._get_parent_ids_by_name = Mock(return_value=parent_ids_by_name)

            task._create_network_member_group = Mock()

            # Execute the test.
            task._process_parent(sobject_type, record_names)

            # Assert scenario execute as expected.

            task._get_parent_ids_by_name.assert_not_called()

            parent_ids_by_name.items.assert_not_called()

            task._create_network_member_group.assert_not_called()

    def test_process_parent__with_names(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        record_names = [
            "Name_0",
            "Name_1",
            "Name_2",
        ]

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        parent_ids_by_name = Mock()
        parent_ids_by_name.items.return_value = [
            ("Name_0", "Id_0"),
            ("Name_1", None),
            ("Name_2", "Id_2"),
        ]

        task._get_parent_ids_by_name = Mock(return_value=parent_ids_by_name)

        task._create_network_member_group = Mock()

        expected_create_network_member_group_calls = []
        for parent_name, parent_id in parent_ids_by_name.items.return_value:
            expected_create_network_member_group_calls.append(
                call(sobject_type, parent_name, parent_id)
            )

        # Execute the test.
        task._process_parent(sobject_type, record_names)

        # Assert scenario execute as expected.

        task._get_parent_ids_by_name.assert_called_once_with(sobject_type, record_names)

        parent_ids_by_name.items.assert_called_once_with()

        task._create_network_member_group.assert_has_calls(
            expected_create_network_member_group_calls
        )

    def test_create_network_member_group__parent_not_found_in_query(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        parent_name = "parent_name"
        parent_id = None

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._parent_ids = set()

        task._network_id = "network_id"

        task.sf = Mock()

        insert_response = Mock()
        task.sf.NetworkMemberGroup.create = Mock(insert_response)

        # Execute the test.
        with self.assertRaises(CumulusCIException) as context:
            task._create_network_member_group(sobject_type, parent_name, parent_id)

        # Assert scenario execute as expected.
        self.assertEqual(
            f'No {sobject_type} record found with Name "{parent_name}"',
            context.exception.args[0],
        )

        task.sf.NetworkMemberGroup.create.assert_not_called()

    def test_create_network_member_group__parent_already_exists(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        parent_name = "parent_name"
        parent_id = "parent_id"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._parent_ids = set()
        task._parent_ids.add(parent_id)

        task._network_id = "network_id"

        task.sf = Mock()
        task.sf.NetworkMemberGroup = Mock()

        insert_response = Mock()
        insert_response.get = Mock()
        task.sf.NetworkMemberGroup.create = Mock(insert_response)

        # Execute the test.
        task._create_network_member_group(sobject_type, parent_name, parent_id)

        task.sf.NetworkMemberGroup.create.assert_not_called()

    def test_create_network_member_group__creating_parent_with_success(self):
        network_name = "network_name"
        sobject_type = "sobject_type"
        parent_name = "parent_name"
        parent_id = "parent_id"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._parent_ids = set()

        task._network_id = "network_id"

        task.sf = Mock()
        task.sf.NetworkMemberGroup = Mock()

        insert_response = {"success": True}

        task.sf.NetworkMemberGroup.create.return_value = insert_response

        # Execute the test.
        task._create_network_member_group(sobject_type, parent_name, parent_id)

        task.sf.NetworkMemberGroup.create.assert_called_once_with(
            {"NetworkId": task._network_id, "ParentId": parent_id}
        )

    def test_create_network_member_group__creating_parent_not_success__with_errors(
        self,
    ):
        network_name = "network_name"
        sobject_type = "sobject_type"
        parent_name = "parent_name"
        parent_id = "parent_id"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._parent_ids = set()

        task._network_id = "network_id"

        task.sf = Mock()

        errors = ["error_0", "error_1"]
        insert_response = {"success": False, "errors": errors}

        task.sf.NetworkMemberGroup.create.return_value = insert_response

        # Execute the test.
        with self.assertRaises(SalesforceException) as context:
            task._create_network_member_group(sobject_type, parent_name, parent_id)

        # Assert scenario execute as expected.
        self.assertEqual(
            f'Error creating NetworkMemberGroup for Network "{task._network_id}" for parent {sobject_type} "{parent_name}" {parent_id}.   Errors: {", ".join(errors)}',
            context.exception.args[0],
        )

        task.sf.NetworkMemberGroup.create.assert_called_once_with(
            {"NetworkId": task._network_id, "ParentId": parent_id}
        )

    def test_create_network_member_group__creating_parent_not_success__no_errors(
        self,
    ):
        network_name = "network_name"
        sobject_type = "sobject_type"
        parent_name = "parent_name"
        parent_id = "parent_id"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._parent_ids = set()

        task._network_id = "network_id"

        task.sf = Mock()

        insert_response = {"success": False, "errors": None}

        task.sf.NetworkMemberGroup.create = Mock(return_value=insert_response)

        # Execute the test.
        with self.assertRaises(SalesforceException) as context:
            task._create_network_member_group(sobject_type, parent_name, parent_id)

        # Assert scenario execute as expected.
        self.assertEqual(
            f'Error creating NetworkMemberGroup for Network "{task._network_id}" for parent {sobject_type} "{parent_name}" {parent_id}.   Errors: {", ".join([])}',
            context.exception.args[0],
        )

        task.sf.NetworkMemberGroup.create.assert_called_once_with(
            {"NetworkId": task._network_id, "ParentId": parent_id}
        )

    def test_run_task__none_profile_names_and_permission_set_names(
        self,
    ):
        network_name = "network_name"

        task = create_task(CreateNetworkMemberGroups, {"network_name": network_name})

        task._get_network_id = Mock(return_value="network_id")
        task._get_network_member_group_parent_ids = Mock(return_value=set(["Id_1"]))

        task._process_parent = Mock()
        expected_process_parent_calls = [
            call("Profile", []),
            call("PermissionSet", []),
        ]

        # Execute the test.
        task._run_task()

        # Assert scenario execute as expected.
        task._get_network_id.assert_called_once_with(task.options.get("network_name"))

        task._get_network_member_group_parent_ids.assert_called_once_with(
            task._get_network_id.return_value
        )

        task._process_parent.assert_has_calls(expected_process_parent_calls)

    def test_run_task__string_profile_names_and_permission_set_names(
        self,
    ):
        network_name = "network_name"
        profile_names = "profile_name"
        permission_set_names = "permission_set_name"

        task = create_task(
            CreateNetworkMemberGroups,
            {
                "network_name": network_name,
                "profile_names": profile_names,
                "permission_set_names": permission_set_names,
            },
        )

        task._get_network_id = Mock(return_value="network_id")
        task._get_network_member_group_parent_ids = Mock(return_value=set(["Id_1"]))
        task._process_parent = Mock()
        expected_process_parent_calls = [
            call("Profile", [profile_names]),
            call("PermissionSet", [permission_set_names]),
        ]

        # Execute the test.
        task._run_task()

        # Assert scenario execute as expected.
        task._get_network_id.assert_called_once_with(task.options.get("network_name"))

        task._get_network_member_group_parent_ids.assert_called_once_with(
            task._get_network_id.return_value
        )

        task._process_parent.assert_has_calls(expected_process_parent_calls)



from contextlib import contextmanager

try:
    from contextlib import nullcontext  # Python 3.7+
except ImportError:

    @contextmanager
    def nullcontext(enter_result=None):
        yield enter_result



import typing as T
from collections import Counter
from contextlib import contextmanager
from itertools import chain, cycle
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Thread
from unittest import mock

import pytest
import yaml
from sqlalchemy import MetaData, create_engine

from cumulusci.core import exceptions as exc
from cumulusci.core.config import OrgConfig
from cumulusci.tasks.bulkdata.delete import DeleteData
from cumulusci.tasks.bulkdata.snowfakery import (
    RunningTotals,
    Snowfakery,
    SnowfakeryWorkingDirectory,
)
from cumulusci.tasks.bulkdata.tests.utils import _make_task
from cumulusci.utils.parallel.task_worker_queues.tests.test_parallel_worker import (
    DelaySpawner,
)

simple_salesforce_yaml = (
    Path(__file__).parent / "snowfakery/simple_snowfakery.recipe.yml"
)
sample_yaml = Path(__file__).parent / "snowfakery/gen_npsp_standard_objects.recipe.yml"
query_yaml = Path(__file__).parent / "snowfakery/query_snowfakery.recipe.yml"

original_refresh_token = OrgConfig.refresh_oauth_token

FAKE_LOAD_RESULTS = (
    {
        "Insert Account": {
            "sobject": "Account",
            "record_type": None,
            "status": "Success",
            "records_processed": 2,
            "total_row_errors": 0,
        },
        "Insert Contact": {
            "sobject": "Contact",
            "record_type": None,
            "status": "Success",
            "records_processed": 2,
            "total_row_errors": 0,
        },
    },
    {
        "Insert Account": {
            "sobject": "Account",
            "record_type": None,
            "status": "Success",
            "records_processed": 3,
            "total_row_errors": 0,
        },
        "Insert Contact": {
            "sobject": "Contact",
            "record_type": None,
            "status": "Success",
            "records_processed": 3,
            "total_row_errors": 0,
        },
    },
)


def table_values(connection, table):
    query = f"select * from {table.name}"
    values = [val for val in connection.execute(query)]
    return values


def call_closure():
    """Simulate a cycle of load results without doing a real load."""
    return_values = cycle(iter(FAKE_LOAD_RESULTS))

    def __call__(self, *args, **kwargs):
        """Like the __call__ of _run_task, but also capture calls
        in a normal mock_values structure."""

        # Manipulating "self" from a mock side-effect is a challenge.
        # So we need a "real function"
        self.return_values = {"step_results": next(return_values)}
        rc = __call__.mock(*args, **kwargs)
        # remember the values that would have been loaded for later inspection in tests
        values_loaded = db_values_from_db_url(self.options["database_url"])
        __call__.mock.mock_calls[-1].values_loaded = values_loaded
        return rc

    __call__.mock = mock.Mock()

    return __call__


def db_values_from_db_url(database_url):
    engine = create_engine(database_url)
    metadata = MetaData(engine)
    metadata.reflect()

    with engine.connect() as connection:
        values = {
            table_name: table_values(connection, table)
            for table_name, table in metadata.tables.items()
            if table_name[-6:] != "sf_ids"
        }
    return values


@pytest.fixture
def mock_load_data(request):
    with mock.patch(
        "cumulusci.tasks.bulkdata.load.LoadData.__call__", call_closure()
    ) as __call__:
        yield __call__.mock


@pytest.fixture
def threads_instead_of_processes(request):
    with mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Process",
        wraps=Thread,
    ) as t:
        yield t


@pytest.fixture
def fake_processes_and_threads(request):
    class FakeProcessManager:
        def __init__(self):
            self.processes = []

        def __call__(self, target, args, daemon):
            res = self.process_handler(target, args, daemon, index=len(self.processes))
            self.processes.append(res)
            return res

    process_manager = FakeProcessManager()

    with mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Thread",
        process_manager,
    ), mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Process",
        process_manager,
    ):
        yield process_manager


@pytest.fixture
def snowfakery(request, create_task):
    def snowfakery(**kwargs):
        return create_task(Snowfakery, kwargs)

    return snowfakery


@contextmanager
def temporary_file_path(filename):
    with TemporaryDirectory() as tmpdirname:
        path = Path(tmpdirname) / filename
        yield path


@pytest.fixture()
def ensure_accounts(create_task, run_code_without_recording, sf):
    """Delete all accounts and create a certain number of new ones"""

    @contextmanager
    def _ensure_accounts(number_of_accounts):
        def setup(number):
            task = create_task(DeleteData, {"objects": "Entitlement, Account"})
            task()
            for i in range(0, number):
                sf.Account.create({"Name": f"Account {i}"})

        run_code_without_recording(lambda: setup(number_of_accounts))
        yield
        run_code_without_recording(lambda: setup(0))

    return _ensure_accounts


class SnowfakeryTaskResults(T.NamedTuple):
    """Results from a Snowfakery data generation process"""

    task: Snowfakery  # The task, so we can inspect its return_values
    outbox: Path  # The working directory, to look at mapping files, DB files, etc.


@pytest.fixture()
def run_snowfakery_and_inspect_mapping(
    run_snowfakery_and_yield_results,
):
    """Run Snowfakery with some defaulted or overriden options.
    Yield a mapping file for inspection that it was the right file.

    Defaults are same as run_snowfakery_and_yield_results.
    """

    def _run_snowfakery_and_inspect_mapping(**options):
        with run_snowfakery_and_yield_results(**options) as results:
            return get_mapping_from_snowfakery_task_results(results)

    return _run_snowfakery_and_inspect_mapping


def get_mapping_from_snowfakery_task_results(results: SnowfakeryTaskResults):
    """Find the shared mapping file and return it."""
    template_dir = SnowfakeryWorkingDirectory(results.outbox / "template_1/")
    temp_mapping = template_dir.mapping_file
    with open(temp_mapping) as f:
        mapping = yaml.safe_load(f)

    other_mapping = Path(str(temp_mapping).replace("template_1", "1_1"))
    # check that it's truly shared
    assert temp_mapping.read_text() == other_mapping.read_text()
    return mapping


def get_record_counts_from_snowfakery_results(
    results: SnowfakeryTaskResults,
):
    """Collate the record counts from Snowfakery outbox directories"""

    rollups = Counter()
    for subdir in results.outbox.glob("*"):
        if "template" not in subdir.name:
            record_counts = SnowfakeryWorkingDirectory(subdir).get_record_counts()
            rollups.update(record_counts)
    return rollups


@pytest.fixture()
def run_snowfakery_and_yield_results(
    snowfakery, threads_instead_of_processes, mock_load_data
):
    @contextmanager
    def _run_snowfakery_and_inspect_mapping_and_example_records(**options):
        with TemporaryDirectory() as workingdir:
            workingdir = Path(workingdir) / "tempdir"
            task = snowfakery(
                run_until_recipe_repeated=2,
                working_directory=workingdir,
                **options,
            )
            task()
            outbox = Path(workingdir) / "data_load_outbox/"
            yield SnowfakeryTaskResults(task, outbox)

    return _run_snowfakery_and_inspect_mapping_and_example_records


class TestSnowfakery:
    def test_no_options(self):
        with pytest.raises(exc.TaskOptionsError, match="recipe"):
            _make_task(Snowfakery, {})

    @mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Process",
    )
    def test_simple_snowfakery(self, Process, mock_load_data, create_task):
        task = create_task(
            Snowfakery,
            {
                "recipe": sample_yaml,
            },
        )
        task()
        assert mock_load_data.mock_calls
        # should not be called for a simple one-rep load
        assert not Process.mock_calls

    @mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Process",
    )
    @pytest.mark.vcr()
    def test_snowfakery_query_salesforce(self, Process, mock_load_data, create_task):
        task = create_task(
            Snowfakery,
            {
                "recipe": query_yaml,
            },
        )
        task()
        assert mock_load_data.mock_calls
        # should not be called for a simple one-rep load
        assert not Process.mock_calls

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 3)
    def test_small(
        self, mock_load_data, threads_instead_of_processes, create_task_fixture
    ):
        task = create_task_fixture(
            Snowfakery,
            {"recipe": sample_yaml, "run_until_recipe_repeated": "7"},
        )
        task()
        # Batch size was 3, so 7 records takes
        # one initial batch plus two parallel batches
        assert len(mock_load_data.mock_calls) == 3, mock_load_data.mock_calls
        # One should be in a sub-process/thread
        assert len(threads_instead_of_processes.mock_calls) == 2

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 3)
    def test_multi_part(
        self, threads_instead_of_processes, mock_load_data, create_task_fixture
    ):
        task = create_task_fixture(
            Snowfakery,
            {"recipe": sample_yaml, "run_until_recipe_repeated": 15},
        )
        task()
        assert (
            len(mock_load_data.mock_calls) > 3
        )  # depends on the details of the tuning
        assert (
            len(threads_instead_of_processes.mock_calls)
            == len(mock_load_data.mock_calls) - 1
        )

    @mock.patch(
        "cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue.WorkerQueue.Process",
    )
    def test_run_until_loaded(
        self, create_subprocess, mock_load_data, create_task_fixture
    ):
        task = create_task_fixture(
            Snowfakery,
            {"recipe": sample_yaml, "run_until_records_loaded": "Account:1"},
        )
        task()
        assert mock_load_data.mock_calls
        # should not be called for a simple one-rep load
        assert not create_subprocess.mock_calls

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 3)
    def test_run_until_loaded_2_parts(
        self, threads_instead_of_processes, mock_load_data, create_task_fixture
    ):
        task = create_task_fixture(
            Snowfakery,
            {"recipe": sample_yaml, "run_until_records_loaded": "Account:6"},
        )
        task()
        assert len(mock_load_data.mock_calls) == 2
        assert len(threads_instead_of_processes.mock_calls) == 1

    # There was previously a failed attempt at testing the connected app here.
    # Could try again after Snowfakery 2.0 launch.
    # https://github.com/SFDO-Tooling/CumulusCI/blob/c7e0d7552394b3ac268cb373ffb24b72b5c059f3/cumulusci/tasks/bulkdata/tests/test_snowfakery.py#L165-L197https://github.com/SFDO-Tooling/CumulusCI/blob/c7e0d7552394b3ac268cb373ffb24b72b5c059f3/cumulusci/tasks/bulkdata/tests/test_snowfakery.py#L165-L197

    @pytest.mark.vcr()
    def test_run_until_records_in_org__none_needed(
        self, threads_instead_of_processes, mock_load_data, create_task, ensure_accounts
    ):
        with ensure_accounts(6):
            task = create_task(
                Snowfakery,
                {"recipe": sample_yaml, "run_until_records_in_org": "Account:6"},
            )
            task()
        assert len(mock_load_data.mock_calls) == 0, mock_load_data.mock_calls
        assert (
            len(threads_instead_of_processes.mock_calls) == 0
        ), threads_instead_of_processes.mock_calls

    @pytest.mark.vcr()
    def test_run_until_records_in_org__one_needed(
        self,
        sf,
        threads_instead_of_processes,
        mock_load_data,
        create_task,
        ensure_accounts,
    ):
        with ensure_accounts(10):
            # org reports 10 records in org
            # so we only need 6 more.
            # That will be one "initial" batch plus one "parallel" batch
            task = create_task(
                Snowfakery,
                {"recipe": sample_yaml, "run_until_records_in_org": "Account:16"},
            )
            task.logger = mock.Mock()
            task()
        print(task.logger.mock_calls)
        assert len(mock_load_data.mock_calls) == 2, mock_load_data.mock_calls
        assert len(threads_instead_of_processes.mock_calls) == 1

    @pytest.mark.vcr()
    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 3)
    def test_run_until_records_in_org__multiple_needed(
        self,
        threads_instead_of_processes,
        mock_load_data,
        snowfakery,
        ensure_accounts,
    ):
        with ensure_accounts(10):
            task = snowfakery(recipe=sample_yaml, run_until_records_in_org="Account:16")
            task()

        assert len(mock_load_data.mock_calls) == 2, mock_load_data.mock_calls
        assert (
            len(threads_instead_of_processes.mock_calls) == 1
        ), threads_instead_of_processes.mock_calls

    def test_inaccessible_generator_yaml(self, snowfakery):
        with pytest.raises(exc.TaskOptionsError, match="recipe"):
            task = snowfakery(
                recipe=sample_yaml / "junk",
            )
            task()

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.get_debug_mode", lambda: True)
    @mock.patch("psutil.cpu_count", lambda logical: 11)
    def test_snowfakery_debug_mode_and_cpu_count(self, snowfakery, mock_load_data):
        task = snowfakery(recipe=sample_yaml, run_until_recipe_repeated="5")
        with mock.patch.object(task, "logger") as logger:
            task()
        assert "Using 11 workers" in str(logger.mock_calls)

    def test_record_count(self, snowfakery, mock_load_data):
        task = snowfakery(recipe="datasets/recipe.yml", run_until_recipe_repeated="4")
        with mock.patch.object(task, "logger") as logger:
            task()
        mock_calls_as_string = str(logger.mock_calls)
        assert "Account: 5 successes" in mock_calls_as_string, mock_calls_as_string[
            -500:
        ]
        assert "Contact: 5 successes" in mock_calls_as_string, mock_calls_as_string[
            -500:
        ]

    def test_run_until_wrong_format(self, snowfakery):
        with pytest.raises(exc.TaskOptionsError, match="Ten"):
            task = snowfakery(
                recipe=sample_yaml, run_until_records_loaded="Account:Ten"
            )
            task()

    def test_run_until_wrong_format__2(self, snowfakery):
        with pytest.raises(exc.TaskOptionsError, match="Ten"):
            task = snowfakery(
                recipe=sample_yaml, run_until_records_loaded="Account_Ten"
            )
            task()

    def test_run_reps_wrong_format(self, snowfakery):
        with pytest.raises(exc.TaskOptionsError, match="Ten"):
            task = snowfakery(recipe=sample_yaml, run_until_recipe_repeated="Ten")
            task()

    def test_run_until_conflicting_params(self, snowfakery):
        with pytest.raises(exc.TaskOptionsError, match="only one of"):
            task = snowfakery(
                recipe=sample_yaml,
                run_until_records_loaded="Account_Ten",
                run_until_recipe_repeated="1",
            )
            task()

    def test_working_directory(self, snowfakery, mock_load_data):
        with TemporaryDirectory() as t:
            working_directory = Path(t) / "junkdir"
            task = snowfakery(
                recipe=sample_yaml,
                run_until_recipe_repeated="1",
                working_directory=str(working_directory),
            )
            task()
            assert (working_directory / "data_load_outbox").exists()

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 1)
    def test_failures_in_subprocesses__last_batch(
        self, snowfakery, mock_load_data, fake_processes_and_threads
    ):
        class FakeProcess(DelaySpawner):
            def __init__(self, target, args, daemon, index):
                super().__init__(target, args, daemon)
                self.counter = 0
                self.task_class = args[0]["task_class"]
                self.index = index
                try:
                    self._finish()
                except AssertionError:
                    pass

            def is_alive(self):
                print("Alive?", self.task_class, self.index, self.counter, self)
                self.counter += 1
                if self.counter > 3:
                    return False
                return True

        fake_processes_and_threads.process_handler = FakeProcess

        class LoadDataSucceedsOnceThenFails:
            count = 0

            def __call__(self):
                self.count += 1
                if self.count > 1:
                    raise AssertionError("XYZZY")

        mock_load_data.side_effect = LoadDataSucceedsOnceThenFails()

        task = snowfakery(
            recipe=sample_yaml,
            run_until_records_loaded="Account:10",
            num_processes=3,  # todo: test this is enforced
        )
        with mock.patch.object(task, "logger") as logger:
            with pytest.raises(exc.BulkDataException):
                task()
        assert "XYZZY" in str(logger.mock_calls)

    def test_running_totals_repr(self):
        r = RunningTotals()
        r.errors = 12
        r.successes = 11
        assert "11" in repr(r)

    def test_generate_mapping_file__loadfile__inferred(
        self, run_snowfakery_and_inspect_mapping
    ):

        mapping = run_snowfakery_and_inspect_mapping(recipe=simple_salesforce_yaml)

        assert mapping["Insert Account"]["api"] == "bulk"
        assert mapping["Insert Contact"].get("bulk_mode") is None
        assert list(mapping.keys()) == ["Insert Account", "Insert Contact"]

    def test_generate_mapping_file__loadfile__overridden(
        self, run_snowfakery_and_inspect_mapping
    ):
        loading_rules = str(simple_salesforce_yaml).replace(
            ".recipe.yml", "_2.load.yml"
        )
        mapping = run_snowfakery_and_inspect_mapping(
            recipe=simple_salesforce_yaml, loading_rules=str(loading_rules)
        )

        assert mapping["Insert Account"].get("api") is None
        assert mapping["Insert Contact"]["bulk_mode"].lower() == "parallel"
        assert list(mapping.keys()) == ["Insert Contact", "Insert Account"]

    def test_generate_mapping_file__loadfile_multiple_files(
        self, run_snowfakery_and_inspect_mapping
    ):
        loading_rules = (
            str(simple_salesforce_yaml).replace(".recipe.yml", "_2.load.yml")
            + ","
            + str(simple_salesforce_yaml).replace(".recipe.yml", ".load.yml")
        )
        mapping = run_snowfakery_and_inspect_mapping(
            recipe=simple_salesforce_yaml, loading_rules=str(loading_rules)
        )

        assert mapping["Insert Account"]["api"] == "bulk"
        assert mapping["Insert Contact"]["bulk_mode"].lower() == "parallel"
        assert list(mapping.keys()) == ["Insert Contact", "Insert Account"]

    def test_options(
        self,
        mock_load_data,
        threads_instead_of_processes,
        run_snowfakery_and_yield_results,
    ):
        options_yaml = str(sample_yaml).replace(
            "gen_npsp_standard_objects.recipe.yml", "options.recipe.yml"
        )
        with run_snowfakery_and_yield_results(
            recipe=options_yaml, recipe_options="xyzzy:7"
        ) as results:
            record_counts = get_record_counts_from_snowfakery_results(results)
        assert record_counts["Account"] == 7

    @mock.patch("cumulusci.tasks.bulkdata.snowfakery.MIN_PORTION_SIZE", 3)
    def test_multi_part_uniqueness(self, mock_load_data, create_task_fixture):
        task = create_task_fixture(
            Snowfakery,
            {
                "recipe": Path(__file__).parent / "snowfakery/unique_values.recipe.yml",
                "run_until_recipe_repeated": 15,
            },
        )
        task()

        all_rows = chain(
            *(call.values_loaded["blah"] for call in mock_load_data.mock_calls)
        )
        unique_values = [row.value for row in all_rows]
        assert len(unique_values) == len(set(unique_values))

        # See also W-10142031

    # def test_generate_mapping_file(self):
    #     with temporary_file_path("mapping.yml") as temp_mapping:
    #         with temp_sqlite_database_url() as database_url:
    #             task = _make_task(
    #                 GenerateDataFromYaml,
    #                 {
    #                     "options": {
    #                         "generator_yaml": sample_yaml,
    #                         "database_url": database_url,
    #                         "generate_mapping_file": temp_mapping,
    #                     }
    #                 },
    #             )
    #             task()
    #         mapping = yaml.safe_load(open(temp_mapping))
    #         assert mapping["Insert Account"]["fields"]

    # def test_use_mapping_file(self):
    #     assert vanilla_mapping_file.exists()
    #     with temp_sqlite_database_url() as database_url:
    #         task = _make_task(
    #             GenerateDataFromYaml,
    #             {
    #                 "options": {
    #                     "generator_yaml": sample_yaml,
    #                     "database_url": database_url,
    #                     "mapping": vanilla_mapping_file,
    #                 }
    #             },
    #         )
    #         task()
    #         self.assertRowsCreated(database_url)

    # def test_num_records(self):
    #     with temp_sqlite_database_url() as database_url:
    #         task = _make_task(
    #             GenerateDataFromYaml,
    #             {
    #                 "options": {
    #                     "generator_yaml": simple_yaml,
    #                     "database_url": database_url,
    #                 }
    #             },
    #         )
    #         task()
    #         assert len(self.assertRowsCreated(database_url)) == 1, len(
    #             self.assertRowsCreated(database_url)
    #         )

    # @mock.patch(
    #     "cumulusci.tasks.bulkdata.generate_and_load_data_from_yaml.GenerateAndLoadDataFromYaml._dataload"
    # )
    # def test_simple_generate_and_load_with_numrecords(self, _dataload):
    #     task = _make_task(s
    #         GenerateAndLoadDataFromYaml,
    #         {
    #             "options": {
    #                 "generator_yaml": simple_yaml,
    #                 "num_records": 11,
    #                 "num_records_tablename": "Account",
    #             }
    #         },
    #     )
    #     task()
    #     assert len(_dataload.mock_calls) == 1

    # @mock.patch(
    #     "cumulusci.tasks.bulkdata.generate_and_load_data_from_yaml.GenerateAndLoadDataFromYaml._dataload"
    # )
    # def test_simple_generate_and_load(self, _dataload):
    #     task = _make_task(
    #         GenerateAndLoadDataFromYaml,
    #         {
    #             "options": {
    #                 "generator_yaml": simple_yaml,
    #                 "num_records": 11,
    #                 "num_records_tablename": "Account",
    #             }
    #         },
    #     )
    #     task()
    #     assert len(_dataload.mock_calls) == 1

    # @mock.patch("cumulusci.tasks.bulkdata.generate_from_yaml.generate_data")
    # def test_exception_handled_cleanly(self, generate_data):
    #     generate_data.side_effect = AssertionError("Foo")
    #     with pytest.raises(AssertionError) as e:
    #         task = _make_task(
    #             GenerateAndLoadDataFromYaml,
    #             {
    #                 "options": {
    #                     "generator_yaml": simple_yaml,
    #                     "num_records": 11,
    #                     "num_records_tablename": "Account",
    #                 }
    #             },
    #         )
    #         task()
    #         assert "Foo" in str(e.value)
    #     assert len(generate_data.mock_calls) == 1

    # @mock.patch(
    #     "cumulusci.tasks.bulkdata.generate_and_load_data_from_yaml.GenerateAndLoadDataFromYaml._dataload"
    # )
    # def test_batching(self, _dataload):
    #     with temp_sqlite_database_url() as database_url:
    #         task = _make_task(
    #             GenerateAndLoadDataFromYaml,
    #             {
    #                 "options": {
    #                     "generator_yaml": simple_yaml,
    #                     "num_records": 14,
    #                     "batch_size": 6,
    #                     "database_url": database_url,
    #                     "num_records_tablename": "Account",
    #                     "data_generation_task": "cumulusci.tasks.bulkdata.generate_from_yaml.GenerateDataFromYaml",
    #                     "reset_oids": False,
    #                 }
    #             },
    #         )
    #         task()
    #         assert len(_dataload.mock_calls) == 3
    #         task = None  # clean up db?

    #         engine = create_engine(database_url)
    #         connection = engine.connect()
    #         records = list(connection.execute("select * from Account"))
    #         connection.close()
    #         assert len(records) == 14 % 6  # leftovers

    # def test_mismatched_options(self):
    #     with pytest.raises(exc.exc.TaskOptionsError) as e:
    #         task = _make_task(
    #             GenerateDataFromYaml,
    #             {"options": {"generator_yaml": sample_yaml, "num_records": 10}},
    #         )
    #         task()
    #     assert "without num_records_tablename" in str(e.exception)

    # def generate_continuation_data(self, fileobj):
    #     g = data_generator_runtime.Globals()
    #     o = data_generator_runtime.ObjectRow(
    #         "Account", {"Name": "Johnston incorporated", "id": 5}
    #     )
    #     g.register_object(o, "The Company", False)
    #     for i in range(0, 5):
    #         # burn through 5 imaginary accounts
    #         g.id_manager.generate_id("Account")
    #     data_generator.save_continuation_yaml(g, fileobj)

    # def test_with_continuation_file(self):
    #     with temp_sqlite_database_url() as database_url:
    #         with temporary_file_path("cont.yml") as continuation_file_path:
    #             with open(continuation_file_path, "w") as continuation_file:
    #                 self.generate_continuation_data(continuation_file)

    #             task = _make_task(
    #                 GenerateDataFromYaml,
    #                 {
    #                     "options": {
    #                         "generator_yaml": sample_yaml,
    #                         "database_url": database_url,
    #                         "mapping": vanilla_mapping_file,
    #                         "continuation_file": continuation_file_path,
    #                     }
    #                 },
    #             )
    #             task()
    #             rows = self.assertRowsCreated(database_url)
    #             assert dict(rows[0])["id"] == 6

    # def test_with_nonexistent_continuation_file(self):
    #     with pytest.raises(exc.TaskOptionsError) as e:
    #         with temp_sqlite_database_url() as database_url:
    #             task = _make_task(
    #                 GenerateDataFromYaml,
    #                 {
    #                     "options": {
    #                         "generator_yaml": sample_yaml,
    #                         "database_url": database_url,
    #                         "mapping": vanilla_mapping_file,
    #                         "continuation_file": "/tmp/foobar/baz/jazz/continuation.yml",
    #                     }
    #                 },
    #             )
    #             task()
    #             rows = self.assertRowsCreated(database_url)
    #             assert dict(rows[0])["id"] == 6

    #     assert "jazz" in str(e.exception)
    #     assert "does not exist" in str(e.exception)

    # def test_generate_continuation_file(self):
    #     with temporary_file_path("cont.yml") as temp_continuation_file:
    #         with temp_sqlite_database_url() as database_url:
    #             task = _make_task(
    #                 GenerateDataFromYaml,
    #                 {
    #                     "options": {
    #                         "generator_yaml": sample_yaml,
    #                         "database_url": database_url,
    #                         "generate_continuation_file": temp_continuation_file,
    #                     }
    #                 },
    #             )
    #             task()
    #         continuation_file = yaml.safe_load(open(temp_continuation_file))
    #         assert continuation_file  # internals of this file are not important to CumulusCI

    # def _run_snowfakery_and_inspect_mapping(self, **options):
    #     with temporary_file_path("mapping.yml") as temp_mapping:
    #         with temp_sqlite_database_url() as database_url:
    #             task = _make_task(
    #                 GenerateDataFromYaml,
    #                 {
    #                     "options": {
    #                         "database_url": database_url,
    #                         "generate_mapping_file": temp_mapping,
    #                         **options,
    #                     }
    #                 },
    #             )
    #             task()
    #         with open(temp_mapping) as f:
    #             mapping = yaml.safe_load(f)
    #     return mapping

    # def test_generate_mapping_file__loadfile_missing(self):
    #     loading_rules = str(simple_snowfakery_yaml).replace(
    #         ".recipe.yml", "_3.load.yml"
    #     )
    #     with pytest.raises(FileNotFoundError):
    #         self._run_snowfakery_and_inspect_mapping(
    #             generator_yaml=simple_snowfakery_yaml, loading_rules=str(loading_rules)
    #         )



import datetime
import os
import re
import unittest

import pytest
import pytz

from cumulusci.core.exceptions import ConfigMergeError, TaskOptionsError
from cumulusci.utils import temporary_dir, touch

from .. import utils


class TestUtils(unittest.TestCase):
    def test_parse_datetime(self):
        dt = utils.parse_datetime("2018-07-30", "%Y-%m-%d")
        self.assertEqual(dt, datetime.datetime(2018, 7, 30, 0, 0, 0, 0, pytz.UTC))

    def test_process_bool_arg(self):
        for arg in (True, "True", "true", "1"):
            self.assertTrue(utils.process_bool_arg(arg))

        for arg in (False, "False", "false", "0"):
            self.assertFalse(utils.process_bool_arg(arg))

        import warnings

        with warnings.catch_warnings(record=True):
            assert utils.process_bool_arg(None) is False

        with pytest.raises(TypeError):
            utils.process_bool_arg(datetime.datetime.now())

        with pytest.raises(TypeError):
            utils.process_bool_arg("xyzzy")

    def test_process_list_arg(self):
        self.assertEqual([1, 2], utils.process_list_arg([1, 2]))
        self.assertEqual(["a", "b"], utils.process_list_arg("a, b"))
        self.assertEqual(None, utils.process_list_arg(None))

    def test_process_glob_list_arg(self):
        with temporary_dir():
            touch("foo.py")
            touch("bar.robot")

            # Expect passing arg as list works.
            self.assertEqual(
                ["foo.py", "bar.robot"],
                utils.process_glob_list_arg(["foo.py", "bar.robot"]),
            )

            # Falsy arg should return an empty list
            self.assertEqual([], utils.process_glob_list_arg(None))
            self.assertEqual([], utils.process_glob_list_arg(""))
            self.assertEqual([], utils.process_glob_list_arg([]))

            # Expect output to be in order given
            self.assertEqual(
                ["foo.py", "bar.robot"],
                utils.process_glob_list_arg("foo.py, bar.robot"),
            )

            # Expect sorted output of glob results
            self.assertEqual(["bar.robot", "foo.py"], utils.process_glob_list_arg("*"))

            # Patterns that don't match any files
            self.assertEqual(
                ["*.bar", "x.y.z"], utils.process_glob_list_arg("*.bar, x.y.z")
            )

            # Recursive
            os.mkdir("subdir")
            filename = os.path.join("subdir", "baz.resource")
            touch(filename)
            self.assertEqual([filename], utils.process_glob_list_arg("**/*.resource"))

    def test_decode_to_unicode(self):
        self.assertEqual("\xfc", utils.decode_to_unicode(b"\xfc"))
        self.assertEqual("\u2603", utils.decode_to_unicode("\u2603"))
        self.assertEqual(None, utils.decode_to_unicode(None))


class TestMergedConfig(unittest.TestCase):
    def test_init(self):
        config = utils.merge_config(
            {
                "universal_config": {"hello": "world"},
                "user_config": {"hello": "christian"},
            }
        )
        self.assertEqual(config["hello"], "christian")

    def test_merge_failure(self):
        with self.assertRaises(ConfigMergeError) as cm:
            utils.merge_config(
                {
                    "universal_config": {"hello": "world", "test": {"sample": 1}},
                    "user_config": {"hello": "christian", "test": [1, 2]},
                }
            )
        exception = cm.exception
        self.assertEqual(exception.config_name, "user_config")


class TestDictMerger(unittest.TestCase):
    """some stuff that didnt get covered by usual usage"""

    def test_merge_into_list(self):
        combo = utils.dictmerge([1, 2], 3)
        self.assertSequenceEqual(combo, [1, 2, 3])

    def test_cant_merge_into_dict(self):
        with self.assertRaises(ConfigMergeError):
            utils.dictmerge({"a": "b"}, 2)

    def test_cant_merge_nonsense(self):
        with self.assertRaises(ConfigMergeError):
            utils.dictmerge(pytz, 2)


class TestProcessListOfPairsDictArg:
    def test_process_list_of_pairs_dict_arg__already_dict(self):
        expected_dict = {"foo": "bar"}
        actual_dict = utils.process_list_of_pairs_dict_arg(expected_dict)
        assert actual_dict is expected_dict

    def test_process_list_of_pairs_dict_arg__valid_values(self):
        valid_values = "foo:bar,baz:boo"
        actual_dict = utils.process_list_of_pairs_dict_arg(valid_values)
        assert actual_dict == {"foo": "bar", "baz": "boo"}

    def test_process_list_of_pairs_dict_arg__uri_values(self):
        uri_value = "companyWebsite:https://www.salesforce.org:8080"
        actual_dict = utils.process_list_of_pairs_dict_arg(uri_value)
        assert actual_dict == {"companyWebsite": "https://www.salesforce.org:8080"}

    def test_process_list_of_pairs_dict_arg__not_dict_or_string(self):
        unsupported = ("foo", "bar")
        error_message = re.escape(
            f"Arg is not a dict or string ({type(unsupported)}): {unsupported}"
        )
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(unsupported)

    def test_process_list_of_pairs_dict_arg__not_name_value_pair(self):
        not_pair = "foo:bar,baz"
        error_message = re.escape("Var is not a name/value pair: baz")
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(not_pair)

    def test_process_list_of_pairs_dict_arg__duplicate_value(self):
        duplicate = "foo:bar,foo:baz"
        error_message = re.escape("Var specified twice: foo")
        with pytest.raises(TaskOptionsError, match=error_message):
            utils.process_list_of_pairs_dict_arg(duplicate)



import shutil
import time
import typing as T
from collections import defaultdict
from contextlib import contextmanager
from datetime import timedelta
from pathlib import Path
from queue import Empty
from tempfile import TemporaryDirectory, mkdtemp

import psutil
import yaml
from snowfakery.api import COUNT_REPS
from sqlalchemy import MetaData, Table, create_engine, func, select

import cumulusci.core.exceptions as exc
from cumulusci.core.config import TaskConfig
from cumulusci.core.debug import get_debug_mode
from cumulusci.core.utils import (
    format_duration,
    process_bool_arg,
    process_list_arg,
    process_list_of_pairs_dict_arg,
)
from cumulusci.tasks.bulkdata.generate_and_load_data_from_yaml import (
    GenerateAndLoadDataFromYaml,
)
from cumulusci.tasks.bulkdata.generate_from_yaml import GenerateDataFromYaml
from cumulusci.tasks.bulkdata.load import LoadData
from cumulusci.tasks.bulkdata.snowfakery_utils.snowfakery_run_until import (
    PortionGenerator,
    determine_run_until,
)
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.utils.parallel.task_worker_queues.parallel_worker_queue import (
    WorkerQueue,
    WorkerQueueConfig,
)

# A portion serves the same process in this system as a "batch" in
# other systems. The term "batch" is not used to avoid confusion with
# Salesforce Bulk API 1.0 batches. For example, a portion of 250_000
# Account records would be broken into roughly 25 Salesforce upload
# batches.

# The system starts at the MIN_PORTION_SIZE and grows towards the
# MAX_PORTION_SIZE. This is to prevent the org wasting time waiting
# for the first portions.
MIN_PORTION_SIZE = 2_000
MAX_PORTION_SIZE = 250_000
ERROR_THRESHOLD = (
    0  # TODO v2.1: Allow this to be a percentage of recent records instead
)

# time between "ticks" where the task re-evaluates its progress
# relatively arbitrary trade-off between busy-waiting and adding latency.
WAIT_TIME = 3

# more loader workers than generators because they spend so much time
# waiting for responses. 4:1 is an experimentally derived ratio
WORKER_TO_LOADER_RATIO = 4

# number of portions we will allow to be on-disk waiting to be loaded
# higher numbers use more disk space.
LOAD_QUEUE_SIZE = 15


class Snowfakery(BaseSalesforceApiTask):

    task_docs = """
    Do a data load with Snowfakery.

    All options are optional.

    The most commonly supplied options are `recipe` and one of the three
    `run_until_...` options.
    """

    task_options = {
        "recipe": {
            "required": True,
            "description": "Path to a Snowfakery recipe file determining what data to generate and load.",
        },
        "run_until_records_in_org": {
            "description": """<sobject>:<count>

      Run the recipe repeatedly until the count of <sobject>
      in the org matches the given <count>.

      For example, `--run_until_records_in_org Account:50_000` means:

      Count the Account records in the org. Let’s say the number
      is 20,000. Thus, we must run the recipe over and
      over again until we generate 30,000 new Account records.
      If the recipe also generates e.g.Contacts, Opportunities or whatever
      else, it generates the appropriate number of them to match.

      Underscores are allowed but optional in big numbers: 2000000
      is the same as 2_000_000.
        """
        },
        "run_until_records_loaded": {
            "description": """<sobject>:<count>

      Run the recipe repeatedly until the number of records of
      <sobject> uploaded in this task execution matches <count>.

      For example, `--run_until_records_loaded Account:50_000` means:

      Run the recipe over and over again
      until we generate 50_000 new Account records. If the recipe
      also generates e.g. Contacts, Opportunities or whatever else, it
      generates the appropriate number of them to match.
        """
        },
        "run_until_recipe_repeated": {
            "description": """Run the recipe <count> times,
            no matter what data is already in the org.

            For example, `--run_until_recipe_repeated 50_000` means
            run the recipe 50_000 times."""
        },
        "working_directory": {"description": "Path for temporary / working files"},
        "loading_rules": {
            "description": "Path to .load.yml file containing rules to use to "
            "load the file. Defaults to `<recipename>.load.yml`. "
            "Multiple files can be comma separated."
        },
        "recipe_options": {
            "required": False,
            "description": """Pass values to override options in the format VAR1:foo,VAR2:bar

             Example: --recipe_options weight:10,color:purple""",
        },
        "bulk_mode": {
            "description": "Set to Serial to force serial mode on all jobs. Parallel is the default."
        },
        "drop_missing_schema": {
            "description": "Set to True to skip any missing objects or fields instead of stopping with an error."
        },
        "num_processes": {
            "description": "Number of data generating processes. Defaults to matching the number of CPUs."
        },
        "ignore_row_errors": {
            "description": "Boolean: should we continue loading even after running into row errors? "
            "Defaults to False."
        },
    }

    def _validate_options(self):
        super()._validate_options()
        # Do not store recipe due to MetaDeploy options freezing
        recipe = self.options.get("recipe")
        recipe = Path(recipe)
        if not recipe.exists():
            raise exc.TaskOptionsError(f"Cannot find recipe `{recipe}`")

        self.num_generator_workers = self.options.get("num_processes", None)
        if self.num_generator_workers is not None:
            self.num_generator_workers = int(self.num_generator_workers)
        self.ignore_row_errors = process_bool_arg(
            self.options.get("ignore_row_errors", False)
        )
        loading_rules = process_list_arg(self.options.get("loading_rules")) or []
        self.loading_rules = [Path(path) for path in loading_rules if path]
        self.recipe_options = process_list_of_pairs_dict_arg(
            self.options.get("recipe_options") or {}
        )

    @property
    def num_loader_workers(self):
        return self.num_generator_workers * WORKER_TO_LOADER_RATIO

    def setup(self):
        self.debug_mode = get_debug_mode()
        if not self.num_generator_workers:
            # logical CPUs do not really improve performance of CPU-bound
            # code, so we ignore them.
            self.num_generator_workers = psutil.cpu_count(logical=False)
            if self.debug_mode:
                self.logger.info(f"Using {self.num_generator_workers} workers")

        self.run_until = determine_run_until(self.options, self.sf)
        self.start_time = time.time()
        self.recipe = Path(self.options.get("recipe"))
        self.job_counter = 0
        self.cached_counts = {}
        self.sobject_counts = defaultdict(RunningTotals)

    ## Todo: Consider when this process runs longer than 2 Hours,
    # what will happen to my sf connection?
    def _run_task(self):
        self.setup()

        portions = PortionGenerator(
            self.run_until.gap,
            MIN_PORTION_SIZE,
            MAX_PORTION_SIZE,
        )

        working_directory = self.options.get("working_directory")
        with self.workingdir_or_tempdir(working_directory) as working_directory:

            self.data_gen_q, self.load_data_q = self.configure_queues(working_directory)
            self.logger.info(f"Working directory is {working_directory}")

            if self.run_until.nothing_to_do:
                self.logger.info(
                    f"Dataload is finished before it started! {self.run_until.nothing_to_do_because}"
                )
                return

            template_path, relevant_sobjects = self._generate_and_load_initial_batch(
                self.load_data_q.outbox_dir
            )

            # disable OrgReordCounts for now until it's reliability can be better
            # tested and documented.

            # Retrieve OrgRecordCounts code from
            # https://github.com/SFDO-Tooling/CumulusCI/commit/7d703c44b94e8b21f165e5538c2249a65da0a9eb#diff-54676811961455410c30d9c9405a8f3b9d12a6222a58db9d55580a2da3cfb870R147

            upload_status = self._loop(
                template_path,
                working_directory,
                None,
                portions,
            )
            self.finish(upload_status, self.data_gen_q, self.load_data_q)

    def configure_queues(self, working_directory):
        """Configure two ParallelWorkerQueues for datagen and dataload"""
        try:
            # in the future, the connected_app should come from the org
            connected_app = self.project_config.keychain.get_service("connected_app")
        except exc.ServiceNotConfigured:  # pragma: no cover
            # to discuss...when can this happen? What are the consequences?
            connected_app = None

        data_gen_q_config = WorkerQueueConfig(
            project_config=self.project_config,
            org_config=self.org_config,
            connected_app=connected_app,
            redirect_logging=True,
            # processes are better for compute-heavy tasks (in Python)
            spawn_class=WorkerQueue.Process,
            parent_dir=working_directory,
            name="data_gen",
            task_class=GenerateDataFromYaml,
            make_task_options=self.data_generator_opts,
            queue_size=0,
            num_workers=self.num_generator_workers,
        )
        data_gen_q = WorkerQueue(data_gen_q_config)

        load_data_q_config = WorkerQueueConfig(
            project_config=self.project_config,
            org_config=self.org_config,
            connected_app=connected_app,
            redirect_logging=True,
            spawn_class=WorkerQueue.Thread,
            parent_dir=working_directory,
            name="data_load",
            task_class=LoadData,
            make_task_options=self.data_loader_opts,
            queue_size=LOAD_QUEUE_SIZE,
            num_workers=self.num_loader_workers,
            rename_directory=self.data_loader_new_directory_name,
        )
        load_data_q = WorkerQueue(load_data_q_config)

        data_gen_q.feeds_data_to(load_data_q)
        return data_gen_q, load_data_q

    def _loop(
        self,
        template_path,
        tempdir,
        org_record_counts_thread,
        portions: PortionGenerator,
    ):
        """The inner loop that controls when data is generated and when we are done."""
        upload_status = self.generate_upload_status(
            portions.next_batch_size,
            template_path,
        )

        while not portions.done(upload_status.total_sets_working_on_or_uploaded):
            if self.debug_mode:
                self.logger.info(f"Working Directory: {tempdir}")

            self.tick(
                upload_status,
                template_path,
                tempdir,
                portions,
            )

            time.sleep(WAIT_TIME)

            upload_status = self._report_status(
                portions.batch_size,
                org_record_counts_thread,
                template_path,
            )

        return upload_status

    def _report_status(
        self,
        batch_size,
        org_record_counts_thread,
        template_path,
    ):
        """Let the user know what is going on."""
        self.logger.info(
            "\n********** PROGRESS *********",
        )

        upload_status = self.generate_upload_status(
            batch_size or 0,
            template_path,
        )

        self.logger.info(upload_status._display(detailed=self.debug_mode))

        if upload_status.sets_failed:
            self.log_failures(
                set(self.load_data_q.failed_job_dirs + self.data_gen_q.failed_job_dirs)
            )

        if upload_status.sets_failed > ERROR_THRESHOLD:
            raise exc.BulkDataException(
                f"Errors exceeded threshold: {upload_status.sets_failed} vs {ERROR_THRESHOLD}"
            )

        # TODO: Retrieve OrgRecordCounts code from
        # https://github.com/SFDO-Tooling/CumulusCI/commit/7d703c44b94e8b21f165e5538c2249a65da0a9eb#diff-54676811961455410c30d9c9405a8f3b9d12a6222a58db9d55580a2da3cfb870R147

        return upload_status

    def update_running_totals(self) -> None:
        while True:
            try:
                results = self.load_data_q.results_reporter.get(block=False)
            except Empty:
                break
            if "results" in results:
                self.update_running_totals_from_load_step_results(results["results"])
            elif "error" in results:
                self.logger.warning(f"Error in load: {results}")
            else:  # pragma: no cover
                self.logger.warning(f"Unexpected message from subtask: {results}")

    def update_running_totals_from_load_step_results(self, results: dict) -> None:
        for result in results["step_results"].values():
            sobject_name = result["sobject"]
            totals = self.sobject_counts[sobject_name]
            totals.errors += result["total_row_errors"]
            totals.successes += result["records_processed"] - result["total_row_errors"]

    def print_running_totals(self):
        for name, result in self.sobject_counts.items():
            self.logger.info(
                f"       {name}: {result.successes:,} successes, {result.errors:,} errors"
            )

    def tick(
        self,
        upload_status,
        template_path,
        tempdir,
        portions: PortionGenerator,
    ):
        """Called every few seconds, to make new data generators if needed."""
        self.data_gen_q.tick()

        if portions.done(
            upload_status.total_sets_working_on_or_uploaded
        ):  # pragma: no cover
            self.logger.info("Finished Generating")
        # queue is full
        elif self.data_gen_q.num_free_workers and self.data_gen_q.full:
            self.logger.info("Waiting before datagen (load queue is full)")
        else:
            for i in range(self.data_gen_q.num_free_workers):
                upload_status = self.generate_upload_status(
                    portions.next_batch_size,
                    template_path,
                )
                self.job_counter += 1
                batch_size = portions.next_batch(
                    upload_status.total_sets_working_on_or_uploaded
                )
                if not batch_size:
                    self.logger.info(
                        "All scheduled portions generated and being uploaded"
                    )
                    break
                job_dir = self.generator_data_dir(
                    self.job_counter, template_path, batch_size, tempdir
                )
                self.data_gen_q.push(job_dir)
                self.update_running_totals()
                self.print_running_totals()

    def finish(self, upload_status, data_gen_q, load_data_q):
        """Wait for jobs to finish"""
        old_message = None
        cooldown = 5
        while data_gen_q.workers + load_data_q.workers:
            datagen_workers = f"{len(data_gen_q.workers)} datagen, "
            msg = f"Waiting for {datagen_workers}{len(load_data_q.workers)} upload workers to finish"
            if old_message != msg or cooldown < 1:
                old_message = msg
                self.logger.info(msg)
                self.update_running_totals()
                self.print_running_totals()
                cooldown = 5
            else:
                cooldown -= 1
            data_gen_q.tick()
            time.sleep(WAIT_TIME)

        self.log_failures(set(load_data_q.failed_job_dirs + data_gen_q.failed_job_dirs))

        self.logger.info("")
        self.logger.info(" == Results == ")
        self.update_running_totals()
        self.print_running_totals()
        elapsed = format_duration(timedelta(seconds=time.time() - self.start_time))

        if self.run_until.sobject_name:
            result_msg = f"{self.sobject_counts[self.run_until.sobject_name].successes} {self.run_until.sobject_name} records and associated records"
        else:
            result_msg = f"{upload_status.target_count:,} iterations"

        self.logger.info(f"☃ Snowfakery created {result_msg} in {elapsed}.")

    def log_failures(self, dirs=T.Sequence[Path]):
        """Log failures from sub-processes to main process"""
        for failure_dir in dirs:
            f = Path(failure_dir) / "exception.txt"
            if not f.exists():  # pragma: no cover
                continue
            exception = f.read_text(encoding="utf-8").strip().split("\n")[-1]
            self.logger.info(exception)

    def data_loader_opts(self, working_dir: Path):
        wd = SnowfakeryWorkingDirectory(working_dir)

        options = {
            "mapping": wd.mapping_file,
            "reset_oids": False,
            "database_url": wd.database_url,
            "set_recently_viewed": False,
            "ignore_row_errors": self.ignore_row_errors,
            # don't need to pass loading_rules because they are merged into mapping
        }
        return options

    # TODO: This method is actually based on the number generated,
    #       because it is called before the load.
    #       If there are row errors, it will drift out of correctness
    #       Code needs to be updated to rename again after load.
    #       Or move away from using directory names for math altogether.
    def data_loader_new_directory_name(self, working_dir: Path):
        """Change the directory name to reflect the true number of sets created."""

        wd = SnowfakeryWorkingDirectory(working_dir)
        key = wd.index
        if key not in self.cached_counts:
            self.cached_counts[key] = wd.get_record_counts()

        if not self.run_until.sobject_name:
            return working_dir

        count = self.cached_counts[key][self.run_until.sobject_name]

        path, _ = str(working_dir).rsplit("_", 1)
        new_working_dir = Path(path + "_" + str(count))
        return new_working_dir

    def generator_data_dir(self, idx, template_path, batch_size, parent_dir):
        """Create a new generator directory with a name based on index and batch_size"""
        assert batch_size > 0
        data_dir = parent_dir / (str(idx) + "_" + str(batch_size))
        shutil.copytree(template_path, data_dir)
        return data_dir

    def data_generator_opts(self, working_dir, *args, **kwargs):
        """Generate task options for a data generator"""
        wd = SnowfakeryWorkingDirectory(working_dir)
        name = Path(working_dir).name
        parts = name.rsplit("_", 1)
        batch_size = int(parts[-1])
        plugin_options = {
            "pid": str(wd.index),
            "big_ids": "True",
        }

        return {
            "generator_yaml": str(self.recipe),
            "database_url": wd.database_url,
            "num_records": batch_size,
            "reset_oids": False,
            "continuation_file": wd.continuation_file,
            "num_records_tablename": self.run_until.sobject_name or COUNT_REPS,
            "vars": self.recipe_options,
            "plugin_options": plugin_options,
        }

    def generate_upload_status(
        self,
        batch_size,
        template_dir,
    ):
        """Combine information from the different data sources into a single "report".

        Useful for debugging but also for making decisions about what to do next."""

        def set_count_from_names(names):
            return sum(int(name.split("_")[1]) for name in names)

        rc = UploadStatus(
            target_count=self.run_until.gap,
            sets_queued_to_be_generated=set_count_from_names(
                self.data_gen_q.queued_jobs
            ),
            sets_being_generated=set_count_from_names(self.data_gen_q.inprogress_jobs),
            sets_queued_for_loading=set_count_from_names(self.load_data_q.queued_jobs),
            # note that these may count as already imported in the org
            sets_being_loaded=set_count_from_names(self.load_data_q.inprogress_jobs),
            min_portion_size=MIN_PORTION_SIZE,
            max_portion_size=MAX_PORTION_SIZE,
            user_max_num_generator_workers=self.num_generator_workers,
            user_max_num_loader_workers=self.num_loader_workers,
            elapsed_seconds=int(time.time() - self.start_time),
            # todo: use row-level result from org load for better accuracy
            sets_finished=set_count_from_names(self.load_data_q.outbox_jobs),
            sets_failed=len(self.load_data_q.failed_jobs),
            batch_size=batch_size,
            inprogress_generator_jobs=len(self.data_gen_q.inprogress_jobs),
            inprogress_loader_jobs=len(self.load_data_q.inprogress_jobs),
            data_gen_free_workers=self.data_gen_q.num_free_workers,
        )
        return rc

    @contextmanager
    def workingdir_or_tempdir(self, working_directory: T.Optional[T.Union[Path, str]]):
        """Make a working directory or a temporary directory, as needed"""
        if working_directory:
            working_directory = Path(working_directory)
            working_directory.mkdir()
            self.logger.info(f"Working Directory {working_directory}")
            yield working_directory
        elif self.debug_mode:
            working_directory = Path(mkdtemp())
            self.logger.info(
                f"Due to debug mode, Working Directory {working_directory} will not be removed"
            )
            yield working_directory
        else:
            with TemporaryDirectory() as tempdir:
                yield Path(tempdir)

    def _generate_and_load_initial_batch(self, working_directory: T.Optional[Path]):
        """Generate a single batch to set up all just_once (singleton) objects"""

        template_dir = Path(working_directory) / "template_1"
        template_dir.mkdir()

        plugin_options = {
            "pid": "0",
            "big_ids": "True",
        }
        results = self._generate_and_load_batch(
            template_dir,
            {
                "generator_yaml": self.options.get("recipe"),
                "num_records": 1,  # smallest possible batch to get to parallelizing fast
                "num_records_tablename": self.run_until.sobject_name or COUNT_REPS,
                "loading_rules": self.loading_rules,
                "vars": self.recipe_options,
                "plugin_options": plugin_options,
            },
        )
        self.update_running_totals_from_load_step_results(results)

        # rename directory to reflect real number of sets created.
        wd = SnowfakeryWorkingDirectory(template_dir)
        new_template_dir = self.data_loader_new_directory_name(template_dir)
        shutil.move(template_dir, new_template_dir)
        template_dir = new_template_dir

        # don't send data tables to child processes. All they
        # care about are ID->OID mappings
        wd = SnowfakeryWorkingDirectory(template_dir)
        self._cleanup_object_tables(*wd.setup_engine())

        return template_dir, wd.relevant_sobjects()

    def _generate_and_load_batch(self, tempdir, options) -> dict:
        options = {
            **options,
            "working_directory": tempdir,
            "set_recently_viewed": False,
            "ignore_row_errors": self.ignore_row_errors,
        }
        subtask_config = TaskConfig({"options": options})
        subtask = GenerateAndLoadDataFromYaml(
            project_config=self.project_config,
            task_config=subtask_config,
            org_config=self.org_config,
            flow=self.flow,
            name=self.name,
            stepnum=self.stepnum,
        )
        subtask()
        return subtask.return_values["load_results"][0]

    def _cleanup_object_tables(self, engine, metadata):
        """Delete all tables that do not relate to id->OID mapping"""
        tables = metadata.tables
        tables_to_drop = [
            table
            for tablename, table in tables.items()
            if not tablename.endswith("sf_ids")
        ]
        if tables_to_drop:
            metadata.drop_all(tables=tables_to_drop)


class UploadStatus(T.NamedTuple):
    """Single "report" of the current status of our processes."""

    batch_size: int
    sets_being_generated: int
    sets_queued_to_be_generated: int
    sets_being_loaded: int
    sets_queued_for_loading: int
    sets_finished: int
    target_count: int
    min_portion_size: int
    max_portion_size: int
    user_max_num_loader_workers: int
    user_max_num_generator_workers: int
    elapsed_seconds: int
    sets_failed: int
    inprogress_generator_jobs: int
    inprogress_loader_jobs: int
    data_gen_free_workers: int

    @property
    def total_in_flight(self):
        return (
            self.sets_queued_to_be_generated
            + self.sets_being_generated
            + self.sets_queued_for_loading
            + self.sets_being_loaded
        )

    @property
    def total_sets_working_on_or_uploaded(
        self,
    ):
        return self.total_in_flight + self.sets_finished

    def _display(self, detailed=False):
        most_important_stats = [
            "target_count",
            "total_sets_working_on_or_uploaded",
            "sets_finished",
            "sets_failed",
        ]

        queue_stats = [
            "inprogress_generator_jobs",
            "inprogress_loader_jobs",
        ]

        def display_stats(keys):
            def format(a):
                return f"{a.replace('_', ' ').title()}: {getattr(self, a):,}"

            return (
                "\n"
                + "\n".join(
                    format(a)
                    for a in keys
                    if not a[0] == "_" and not callable(getattr(self, a))
                )
                + "\n"
            )

        rc = display_stats(most_important_stats)
        rc += "\n   ** Workers **\n"
        rc += display_stats(queue_stats)
        if detailed:
            rc += "\n   ** Internals **\n"
            rc += display_stats(
                set(dir(self)) - (set(most_important_stats) & set(queue_stats))
            )
        return rc


class SnowfakeryWorkingDirectory:
    """Helper functions based on well-known filenames in CCI/Snowfakery working directories."""

    def __init__(self, working_dir):
        self.path = working_dir
        self.mapping_file = working_dir / "temp_mapping.yml"
        self.database_file = working_dir / "generated_data.db"
        assert self.mapping_file.exists(), self.mapping_file
        assert self.database_file.exists(), self.database_file
        self.database_url = f"sqlite:///{self.database_file}"
        self.continuation_file = f"{working_dir}/continuation.yml"

    def setup_engine(self):
        """Set up the database engine"""
        engine = create_engine(self.database_url)

        metadata = MetaData(engine)
        metadata.reflect()
        return engine, metadata

    @property
    def index(self) -> str:
        return self.path.name.rsplit("_")[0]

    def get_record_counts(self):
        """Get record counts generated for this portion."""
        engine, metadata = self.setup_engine()

        with engine.connect() as connection:
            record_counts = {
                table_name: self._record_count_from_db(connection, table)
                for table_name, table in metadata.tables.items()
                if table_name[-6:] != "sf_ids"
            }
        # note that the template has its contents deleted so if the cache
        # is ever removed, it will start to return {}
        assert record_counts
        return record_counts

    def _record_count_from_db(self, connection, table: Table):
        """Count rows in a table"""
        stmt = select(func.count()).select_from(table)
        result = connection.execute(stmt)
        return next(result)[0]

    def relevant_sobjects(self):
        with open(self.mapping_file, encoding="utf-8") as f:
            mapping = yaml.safe_load(f)
            return [m.get("sf_object") for m in mapping.values() if m.get("sf_object")]


class RunningTotals:
    errors: int = 0
    successes: int = 0

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.__dict__}>"



import datetime
import json
from unittest import mock

import pytest
from simple_salesforce import SalesforceMalformedRequest

from cumulusci.tasks.push.push_api import (
    BasePushApiObject,
    MetadataPackage,
    MetadataPackageVersion,
    PackagePushError,
    PackagePushJob,
    PackagePushRequest,
    PackageSubscriber,
    SalesforcePushApi,
    batch_list,
)

NAME = "Chewbacca"
SF_ID = "033xxxxxxxxx"
PUSH_API = "push_api"
NAMESPACE = "namespace"
ORG_KEY = "bar"


@pytest.fixture
def metadata_package_versions(metadata_package):
    return [
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="3",
            minor="1",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="2",
            minor="1",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="1",
            minor="1",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="1",
            minor="1",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="1",
            minor="89",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="89",
            minor="1",
            patch="1",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="1",
            minor="1",
            patch="89",
            build="1",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="1",
            minor="1",
            patch="1",
            build="89",
        ),
        MetadataPackageVersion(
            push_api=PUSH_API,
            package=metadata_package,
            name=NAME,
            sf_id=SF_ID,
            state="Beta",
            major="4",
            minor="3",
            patch="1",
            build="1",
        ),
    ]


@pytest.fixture
def sf_push_api():
    return SalesforcePushApi(sf=mock.Mock(), logger=mock.Mock())


@pytest.fixture
def metadata_package():
    return MetadataPackage(
        push_api=mock.MagicMock(), name=NAME, sf_id=SF_ID, namespace=NAMESPACE
    )


@pytest.fixture
def metadata_package_version(metadata_package):
    return MetadataPackageVersion(
        push_api=mock.MagicMock(),
        package=metadata_package,
        name=NAME,
        sf_id=SF_ID,
        state="Beta",
        major="4",
        minor="3",
        patch="2",
        build="1",
    )


@pytest.fixture
def package_push_job():
    return PackagePushJob(
        push_api=mock.MagicMock(),
        request="",
        org="00DS0000003TJJ6MAO",
        status="Succeeded",
        sf_id=SF_ID,
    )


@pytest.fixture
def package_subscriber():
    return PackageSubscriber(
        push_api=mock.MagicMock(),
        version="1.2.3",
        status="Succeeded",
        org_name="foo",
        org_key="bar",
        org_status="Succeeded",
        org_type="Sandbox",
        sf_id=SF_ID,
    )


@pytest.fixture
def package_push_error():
    return PackagePushError(
        push_api="foo",
        sf_id=SF_ID,
        job="Foo",
        severity="high",
        error_type="bar",
        title="foo_bar",
        message="The foo hit the bar",
        details="foo bar, foo, foo bar",
    )


@pytest.fixture
def package_push_request():
    return PackagePushRequest(
        push_api=mock.MagicMock(),
        version="1.2.3",
        start_time="12:03",
        status="Succeeded",
        sf_id=SF_ID,
    )


def test_base_push_format_where():
    base_obj = BasePushApiObject()
    field_name = "id_field"
    sf_id = "006000000XXX000"
    where_clause = "id=001000000XXX000"
    base_obj.sf_id = sf_id

    returned = base_obj.format_where(field_name, where_clause)
    assert "{} = '{}' AND ({})".format(field_name, sf_id, where_clause) == returned

    returned = base_obj.format_where(field_name, None)
    assert "{} = '{}'".format(field_name, sf_id) == returned


def test_metadata_package_init():

    package = MetadataPackage(PUSH_API, NAME)
    assert package.push_api == PUSH_API
    assert package.sf_id is None
    assert package.name == NAME
    assert package.namespace is None

    package = MetadataPackage(PUSH_API, NAME, SF_ID, NAMESPACE)
    assert package.push_api == PUSH_API
    assert package.sf_id == SF_ID
    assert package.name == NAME
    assert package.namespace == NAMESPACE


def test_metadata_package_get_versions(metadata_package):
    expected = f"MetadataPackageId = '{SF_ID}'"
    metadata_package.get_package_versions()
    metadata_package.push_api.get_package_versions.assert_called_once_with(
        expected, None
    )


def test_metadata_package_get_version_objs(metadata_package):
    expected = f"MetadataPackageId = '{SF_ID}'"
    metadata_package.get_package_version_objs()
    metadata_package.push_api.get_package_version_objs.assert_called_once_with(
        expected, None
    )


def test_metadata_package_get_versions_by_id(metadata_package):
    expected = f"MetadataPackageId = '{SF_ID}'"
    metadata_package.get_package_versions_by_id()
    metadata_package.push_api.get_package_versions_by_id.assert_called_once_with(
        expected, None
    )


def test_metadata_package_version_version_number(metadata_package_version):
    expected = "4.3.2 (Beta 1)"
    actual = metadata_package_version.version_number
    assert expected == actual


def test_sf_push_return_query_records(sf_push_api):
    query = "SELECT Id FROM Account"
    records = ["record 1", "record 2", "record 3"]
    results = {"totalSize": 10, "records": records}

    sf_push_api.sf.query_all.return_value = results
    returned = sf_push_api.return_query_records(query)
    assert len(records) == len(returned)

    results["totalSize"] = 0
    sf_push_api.sf.query_all.return_value = results
    returned = sf_push_api.return_query_records(query)
    assert [] == returned


def test_sf_push_format_where(sf_push_api):
    returned = sf_push_api.format_where_clause(None)
    assert "" == returned

    default_where = "Id='001000000XXX000'"
    sf_push_api.default_where = {"Account": default_where}
    returned = sf_push_api.format_where_clause(None, "Object__c")
    assert "" == returned

    returned = sf_push_api.format_where_clause(None, "Account")
    assert " WHERE ({})".format(default_where) == returned

    where = "IsDeleted=False"
    returned = sf_push_api.format_where_clause(where)
    assert " WHERE {}".format(where) == returned
    # No default where for Object__C
    returned = sf_push_api.format_where_clause(where, "Object__c")
    assert " WHERE {}".format(where) == returned

    returned = sf_push_api.format_where_clause(where, "Account")
    assert " WHERE ({}) AND ({})".format(default_where, where) == returned


def test_sf_push_add_query_limit(sf_push_api):
    query = "SELECT Id FROM Account"
    limit = 100
    returned = sf_push_api.add_query_limit(query, limit)
    assert "{} LIMIT {}".format(query, limit) == returned


def test_sf_push_add_query_no_limit(sf_push_api):
    query = "SELECT Id FROM Account"
    returned = sf_push_api.add_query_limit(query, None)
    assert f"{query}" == returned


def test_sf_push_get_packages(sf_push_api):
    query = "SELECT id, name, namespaceprefix FROM MetadataPackage WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_packages("Name='foo'", None)
    sf_push_api.return_query_records.assert_called_once_with(query)


def test_sf_push_get_package_objs(sf_push_api, metadata_package):
    sf_push_api.get_packages = mock.MagicMock()
    packages = {
        "Id": metadata_package.sf_id,
        "Name": metadata_package.name,
        "NamespacePrefix": metadata_package.namespace,
    }
    sf_push_api.get_packages.return_value = [packages]
    actual_result_list = sf_push_api.get_package_objs("Name='foo'", None)
    assert len(actual_result_list) == 1
    actual_result = actual_result_list[0]
    assert packages["Id"] == actual_result.sf_id
    assert packages["Name"] == actual_result.name
    assert packages["NamespacePrefix"] == actual_result.namespace


def test_sf_push_get_packages_by_id(sf_push_api, metadata_package):
    sf_push_api.get_package_objs = mock.MagicMock()
    sf_push_api.get_package_objs.return_value = [metadata_package]
    package_expected = {metadata_package.sf_id: metadata_package}
    package_result = sf_push_api.get_packages_by_id("Name='foo'", None)
    sf_push_api.get_package_objs.assert_called_with("Name='foo'", None)
    assert package_expected == package_result


def test_sf_push_get_package_versions(sf_push_api):
    query = "SELECT Id, Name, MetadataPackageId, ReleaseState, MajorVersion, MinorVersion, PatchVersion, BuildNumber FROM MetadataPackageVersion WHERE Name='foo' ORDER BY MajorVersion DESC, MinorVersion DESC, PatchVersion, BuildNumber DESC"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_package_versions("Name='foo'", None)
    sf_push_api.return_query_records.assert_called_once_with(query)


def test_sf_push_get_package_version_objs(sf_push_api):
    query = "SELECT Id, Name, MetadataPackageId, ReleaseState, MajorVersion, MinorVersion, PatchVersion, BuildNumber FROM MetadataPackageVersion WHERE Name='foo' ORDER BY MajorVersion DESC, MinorVersion DESC, PatchVersion, BuildNumber DESC"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_package_version_objs("Name='foo'", None)
    sf_push_api.return_query_records.assert_called_with(query)


def test_sf_push_get_package_version_by_id(sf_push_api, metadata_package_version):
    sf_push_api.get_package_version_objs = mock.MagicMock()
    sf_push_api.get_package_version_objs.return_value = [metadata_package_version]
    package_expected = {metadata_package_version.sf_id: metadata_package_version}
    package_result = sf_push_api.get_package_versions_by_id("Name='foo'", None)
    sf_push_api.get_package_version_objs.assert_called_with("Name='foo'", None)
    assert package_expected == package_result


def test_sf_push_get_subscribers(sf_push_api):
    query = "SELECT Id, MetadataPackageVersionId, InstalledStatus, InstanceName, OrgName, OrgKey, OrgStatus, OrgType from PackageSubscriber WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    field_names = [
        "Id",
        "MetadataPackageVersionId",
        "InstalledStatus",
        "InstanceName",
        "OrgName",
        "OrgKey",
        "OrgStatus",
        "OrgType",
    ]
    sf_push_api.get_subscribers("Name='foo'", None)
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject="PackageSubscriber"
    )


def test_sf_push_get_subscriber_objs(sf_push_api):
    query = "SELECT Id, MetadataPackageVersionId, InstalledStatus, InstanceName, OrgName, OrgKey, OrgStatus, OrgType from PackageSubscriber WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_subscriber_objs("Name='foo'", None)
    field_names = [
        "Id",
        "MetadataPackageVersionId",
        "InstalledStatus",
        "InstanceName",
        "OrgName",
        "OrgKey",
        "OrgStatus",
        "OrgType",
    ]
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject="PackageSubscriber"
    )


def test_sf_push_get_subscribers_by_org_key(sf_push_api, package_subscriber):
    sf_push_api.get_subscriber_objs = mock.MagicMock()
    sf_push_api.get_subscriber_objs.return_value = [package_subscriber]
    package_expected = {package_subscriber.org_key: package_subscriber}
    package_result = sf_push_api.get_subscribers_by_org_key("Name='foo'", None)
    sf_push_api.get_subscriber_objs.assert_called_with("Name='foo'", None)
    assert package_expected == package_result


def test_sf_push_get_push_requests(sf_push_api):
    query = "SELECT Id, PackageVersionId, ScheduledStartTime, Status FROM PackagePushRequest WHERE Name='foo' ORDER BY ScheduledStartTime DESC"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_push_requests("Name='foo'", None)
    field_names = ["Id", "PackageVersionId", "ScheduledStartTime", "Status"]
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject="PackagePushRequest"
    )


def test_sf_push_get_push_request_objs(sf_push_api):
    query = "SELECT Id, PackageVersionId, ScheduledStartTime, Status FROM PackagePushRequest WHERE Name='foo' ORDER BY ScheduledStartTime DESC"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_push_request_objs("Name='foo'", None)
    field_names = ["Id", "PackageVersionId", "ScheduledStartTime", "Status"]
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject="PackagePushRequest"
    )


def test_sf_push_get_push_requests_by_id(sf_push_api, package_push_request):
    sf_push_api.get_push_request_objs = mock.MagicMock()
    sf_push_api.get_push_request_objs.return_value = [package_push_request]
    package_expected = {package_push_request.sf_id: package_push_request}
    package_result = sf_push_api.get_push_requests_by_id("Name='foo'", None)
    sf_push_api.get_push_request_objs.assert_called_with("Name='foo'", None)
    assert package_expected == package_result


def test_sf_push_get_push_jobs(sf_push_api):
    query = "SELECT Id, PackagePushRequestId, SubscriberOrganizationKey, Status FROM PackagePushJob WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_push_jobs("Name='foo'", None)
    sobject = "PackagePushJob"
    field_names = ["Id", "PackagePushRequestId", "SubscriberOrganizationKey", "Status"]
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject=sobject
    )


def test_sf_push_get_push_job_objs(sf_push_api):
    sobject = "PackagePushJob"
    field_names = ["Id", "PackagePushRequestId", "SubscriberOrganizationKey", "Status"]
    query = f"SELECT Id, PackagePushRequestId, SubscriberOrganizationKey, Status FROM {sobject} WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_push_job_objs("Name='foo'", None)
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject=sobject
    )


def test_sf_push_get_push_jobs_by_id(sf_push_api, package_push_job):
    sf_push_api.get_push_job_objs = mock.MagicMock()
    sf_push_api.get_push_job_objs.return_value = [package_push_job]
    package_expected = {package_push_job.sf_id: package_push_job}
    package_result = sf_push_api.get_push_jobs_by_id("Name='foo'", None)
    sf_push_api.get_push_job_objs.assert_called_with("Name='foo'", None)
    assert package_expected == package_result


def test_sf_push_get_push_errors(sf_push_api):
    query = "SELECT Id, PackagePushJobId, ErrorSeverity, ErrorType, ErrorTitle, ErrorMessage, ErrorDetails FROM PackagePushError WHERE Name='foo'"
    sf_push_api.return_query_records = mock.MagicMock()
    sf_push_api.get_push_errors("Name='foo'", None)
    field_names = [
        "Id",
        "PackagePushJobId",
        "ErrorSeverity",
        "ErrorType",
        "ErrorTitle",
        "ErrorMessage",
        "ErrorDetails",
    ]
    sobject = "PackagePushError"
    sf_push_api.return_query_records.assert_called_with(
        query, field_names=field_names, sobject=sobject
    )


def test_sf_push_get_push_error_objs(sf_push_api, package_push_job, package_push_error):
    sf_push_api.get_push_job_objs = mock.MagicMock()
    sf_push_api.get_push_job_objs.return_value = [package_push_job]
    sf_push_api.lazy = ["jobs"]
    sf_push_api.get_push_errors = mock.MagicMock()
    record = {
        "ErrorSeverity": "high",
        "ErrorType": "bar",
        "ErrorTitle": "foo_bar",
        "ErrorMessage": "The foo hit the bar",
        "ErrorDetails": "foo bar, foo, foo bar",
        "Id": SF_ID,
        "PackagePushJobId": "pkg_push_id",
    }
    sf_push_api.get_push_errors.return_value = [record]

    actual_result_list = sf_push_api.get_push_error_objs("Name='foo'", None)
    sf_push_api.get_push_job_objs.assert_called_once_with(where="Id = 'pkg_push_id'")
    assert len(actual_result_list) == 1
    actual_result = actual_result_list[0]
    assert record["ErrorMessage"] == actual_result.message
    assert record["ErrorDetails"] == actual_result.details
    assert record["Id"] == actual_result.sf_id
    assert actual_result.job == package_push_job


def test_sf_push_get_push_errors_by_id(sf_push_api, package_push_error):
    sf_push_api.get_push_error_objs = mock.MagicMock()
    sf_push_api.get_push_error_objs.return_value = [package_push_error]
    push_error_expected = {package_push_error.sf_id: package_push_error}
    push_error_result = sf_push_api.get_push_errors_by_id("Name='foo'", None)
    sf_push_api.get_push_error_objs.assert_called_with("Name='foo'", None)
    assert push_error_expected == push_error_result


def test_sf_push_cancel_push_request(sf_push_api):
    ref_id = "12"
    sf_push_api.cancel_push_request(ref_id)
    sf_push_api.sf.PackagePushRequest.update.assert_called_once_with(
        ref_id, {"Status": "Canceled"}
    )


def test_sf_push_run_push_request(sf_push_api):
    ref_id = "12"
    sf_push_api.run_push_request(ref_id)
    sf_push_api.sf.PackagePushRequest.update.assert_called_once_with(
        ref_id, {"Status": "Pending"}
    )


def test_sf_push_create_push_request(sf_push_api, metadata_package_version):
    sf_push_api.batch_size = 1
    push_request_id = "0DV?xxxxxx?"
    version_id = metadata_package_version.sf_id = "0KM?xxxxx?"
    orgs = ["00D000000001", "00D000000002"]
    batch_0, batch_1 = [orgs[0]], [orgs[1]]
    start_time = datetime.datetime.now()

    sf_push_api.sf.PackagePushRequest.create.return_value = {"id": push_request_id}
    sf_push_api.sf.base_url = "url"
    sf_push_api._add_batch = mock.MagicMock(side_effect=[batch_0, batch_1])

    actual_id, actual_org_count = sf_push_api.create_push_request(
        metadata_package_version.sf_id, orgs, start_time
    )

    sf_push_api.sf.PackagePushRequest.create.assert_called_once_with(
        {
            "PackageVersionId": version_id,
            "ScheduledStartTime": start_time.isoformat(timespec="seconds"),
        }
    )
    assert mock.call(batch_0, push_request_id) in sf_push_api._add_batch.call_args_list
    assert mock.call(batch_1, push_request_id) in sf_push_api._add_batch.call_args_list
    assert push_request_id == actual_id
    assert 2 == actual_org_count


def test_sf_push_add_push_batch(sf_push_api, metadata_package_version):
    push_request_id = "0DV?xxxxxx?"
    metadata_package_version.sf_id = "0KM?xxxxx?"
    orgs = ["00D000000001", "00D000000002"]
    expected_records_json = json.dumps(
        {
            "records": [
                {
                    "attributes": {"type": "PackagePushJob", "referenceId": orgs[0]},
                    "PackagePushRequestId": push_request_id,
                    "SubscriberOrganizationKey": orgs[0],
                },
                {
                    "attributes": {"type": "PackagePushJob", "referenceId": orgs[1]},
                    "PackagePushRequestId": push_request_id,
                    "SubscriberOrganizationKey": orgs[1],
                },
            ]
        }
    )

    sf_push_api.sf.base_url = "base_url/"

    returned_batch = sf_push_api._add_batch(orgs, push_request_id)
    sf_push_api.sf._call_salesforce.assert_called_once_with(
        "POST", "base_url/composite/tree/PackagePushJob", data=expected_records_json
    )
    assert ["00D000000001", "00D000000002"] == returned_batch


def test_sf_push_add_push_batch_retry(sf_push_api, metadata_package_version):
    push_request_id = "0DV?xxxxxx?"
    orgs = ["00D000000001", "00D000000002", "00D000000003"]
    retry_response = {
        "results": [
            {
                "referenceId": orgs[0],
                "errors": [
                    {"message": "Something bad has happened! Whatever could it be?"}
                ],
            }
        ]
    }
    duplicate_response = {
        "results": [
            {
                "referenceId": orgs[1],
                "errors": [{"message": "", "statusCode": "DUPLICATE_VALUE"}],
            }
        ]
    }
    invalid_response = {
        "results": [
            {
                "referenceId": orgs[2],
                "errors": [{"message": "", "statusCode": "INVALID_OPERATION"}],
            }
        ]
    }

    sf_push_api.sf.base_url = "base_url/"
    sf_push_api.sf._call_salesforce.side_effect = [
        SalesforceMalformedRequest(
            "base_url/composite/tree/PackagePushJob",
            400,
            "resource_name",
            retry_response,
        ),
        SalesforceMalformedRequest(
            "base_url/composite/tree/PackagePushJob",
            400,
            "resource_name",
            duplicate_response,
        ),
        SalesforceMalformedRequest(
            "base_url/composite/tree/PackagePushJob",
            400,
            "resource_name",
            invalid_response,
        ),
        [],
    ]

    returned_batch = sf_push_api._add_batch(orgs, push_request_id)

    assert [orgs[0]] == returned_batch  # only remaining org should be retry-able
    assert 4 == sf_push_api.sf._call_salesforce.call_count


def test_push_batch_list():
    data = ["zero", "one", "two", "three"]

    actual_batch_list = batch_list(data, 1)
    expected_batch_list = [["zero"], ["one"], ["two"], ["three"]]
    assert expected_batch_list == actual_batch_list

    actual_batch_list = batch_list(data, 2)
    expected_batch_list = [["zero", "one"], ["two", "three"]]
    assert expected_batch_list == actual_batch_list

    actual_batch_list = batch_list(data, 3)
    expected_batch_list = [["zero", "one", "two"], ["three"]]
    assert expected_batch_list == actual_batch_list

    actual_batch_list = batch_list(data, 4)
    expected_batch_list = [["zero", "one", "two", "three"]]
    assert expected_batch_list == actual_batch_list

    actual_batch_list = batch_list(data, 5)
    assert expected_batch_list == actual_batch_list

    actual_batch_list = batch_list([], 2)
    expected_batch_list = []
    assert expected_batch_list == actual_batch_list


def test_version_init(metadata_package):
    package = MetadataPackageVersion(
        push_api=PUSH_API,
        package=metadata_package,
        name=NAME,
        sf_id=SF_ID,
        state="Beta",
        major="4",
        minor="3",
        patch="2",
        build="1",
    )
    assert package.push_api == PUSH_API
    assert package.package == metadata_package
    assert package.name == NAME
    assert package.sf_id == SF_ID
    assert package.state == "Beta"
    assert package.major == "4"
    assert package.minor == "3"
    assert package.patch == "2"
    assert package.build == "1"


def test_version_number(metadata_package_version):
    actual = metadata_package_version.version_number
    expected = "4.3.2 (Beta 1)"
    assert actual == expected


def test_metadata_package_get_subscribers(metadata_package_version):
    expected = f"MetadataPackageVersionId = '{SF_ID}'"
    metadata_package_version.get_subscribers()
    metadata_package_version.push_api.get_subscribers.assert_called_once_with(
        expected, None
    )


def test_metadata_package_get_subscriber_objects(metadata_package_version):
    expected = f"MetadataPackageVersionId = '{SF_ID}'"
    metadata_package_version.get_subscriber_objs()
    metadata_package_version.push_api.get_subscriber_objs.assert_called_once_with(
        expected, None
    )


def test_metadata_package_get_subscribers_by_org_key(metadata_package_version):
    expected = f"MetadataPackageVersionId = '{SF_ID}'"
    metadata_package_version.get_subscribers_by_org_key()
    metadata_package_version.push_api.get_subscribers_by_org_key.assert_called_once_with(
        expected, None
    )


def test_metadata_package_push_requests(metadata_package_version):
    expected = f"PackageVersionId = '{SF_ID}'"
    metadata_package_version.get_push_requests()
    metadata_package_version.push_api.get_push_requests.assert_called_once_with(
        expected, None
    )


def test_metadata_package_push_request_objs(metadata_package_version):
    expected = f"PackageVersionId = '{SF_ID}'"
    metadata_package_version.get_push_request_objs()
    metadata_package_version.push_api.get_push_request_objs.assert_called_once_with(
        expected, None
    )


def test_metadata_package_push_requests_by_id(metadata_package_version):
    expected = f"PackageVersionId = '{SF_ID}'"
    metadata_package_version.get_push_requests_by_id()
    metadata_package_version.push_api.get_push_requests_by_id.assert_called_once_with(
        expected, None
    )


def test_version_get_newer_query(metadata_package_version):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND (MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released')"
    metadata_package_version.get_newer_released_version_objs()
    metadata_package_version.package.push_api.get_package_version_objs.assert_called_once_with(
        expected, None
    )


def test_version_get_older_query(metadata_package_version):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND (MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released')"
    metadata_package_version.get_older_released_version_objs()
    metadata_package_version.package.push_api.get_package_version_objs.assert_called_once_with(
        expected, None
    )


def test_version_min_version_query(metadata_package_version, metadata_package):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND (MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released')"
    greater_than = MetadataPackageVersion(
        push_api=PUSH_API,
        package=metadata_package,
        name=NAME,
        sf_id=SF_ID,
        state="Beta",
        major="2",
        minor="2",
        patch="2",
        build="1",
    )
    metadata_package_version.get_older_released_version_objs(greater_than)
    metadata_package_version.package.push_api.get_package_version_objs.assert_called_once_with(
        expected, None
    )


def test_version_min_version_query_integration(
    metadata_package_version, metadata_package_versions, metadata_package
):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND (MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released')"
    min_version = MetadataPackageVersion(
        push_api=PUSH_API,
        package=metadata_package,
        name=NAME,
        sf_id=SF_ID,
        state="Beta",
        major="2",
        minor="1",
        patch="1",
        build="1",
    )
    metadata_package_version.package.push_api.get_package_version_objs.return_value = (
        metadata_package_versions
    )
    assert (
        len(metadata_package_version.get_older_released_version_objs(min_version)) == 3
    )
    metadata_package_version.package.push_api.get_package_version_objs.assert_called_once_with(
        expected, None
    )


def test_version_get_newer(metadata_package_version):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released'"  # AND (MajorVersion > 1 OR (MajorVersion = 1 AND MinorVersion > 2) OR (MajorVersion = 1 AND MinorVersion = 2 AND PatchVersion >= 3))"
    metadata_package_version.package.get_package_version_objs = mock.MagicMock()
    metadata_package_version.get_newer_released_version_objs()
    metadata_package_version.package.get_package_version_objs.assert_called_once_with(
        expected
    )


def test_version_get_older(
    metadata_package_version, metadata_package_versions, metadata_package
):
    expected = "MetadataPackageId = '033xxxxxxxxx' AND ReleaseState = 'Released'"
    metadata_package_version.package.get_package_version_objs = mock.MagicMock()
    metadata_package_version.package.get_package_version_objs.return_value = (
        metadata_package_versions
    )
    min_version = MetadataPackageVersion(
        push_api=PUSH_API,
        package=metadata_package,
        name=NAME,
        sf_id=SF_ID,
        state="Beta",
        major="1",
        minor="1",
        patch="1",
        build="1",
    )
    assert (
        len(metadata_package_version.get_older_released_version_objs(min_version)) == 8
    )
    metadata_package_version.package.get_package_version_objs.assert_called_once_with(
        expected
    )


def test_package_push_job_get_push_errors(package_push_job):
    expected = f"PackagePushJobId = '{SF_ID}'"
    package_push_job.get_push_errors()
    package_push_job.push_api.get_push_errors.assert_called_once_with(expected, None)


def test_package_push_job_get_push_error_objects(package_push_job):
    expected = f"PackagePushJobId = '{SF_ID}'"
    package_push_job.get_push_error_objs()
    package_push_job.push_api.get_push_error_objs.assert_called_once_with(
        expected, None
    )


def test_package_push_job_get_push_errors_by_id(package_push_job):
    expected = f"PackagePushJobId = '{SF_ID}'"
    package_push_job.get_push_errors_by_id()
    package_push_job.push_api.get_push_errors_by_id.assert_called_once_with(
        expected, None
    )


def test_package_push_errors(package_push_error):
    assert package_push_error.push_api == "foo"
    assert package_push_error.sf_id == SF_ID
    assert package_push_error.job == "Foo"
    assert package_push_error.severity == "high"

    assert package_push_error.error_type == "bar"
    assert package_push_error.title == "foo_bar"
    assert package_push_error.message == "The foo hit the bar"
    assert package_push_error.details == "foo bar, foo, foo bar"


def test_package_push_request_get_push_jobs(package_push_request):
    expected = f"PackagePushRequestId = '{SF_ID}'"
    package_push_request.get_push_jobs()
    package_push_request.push_api.get_push_jobs.assert_called_once_with(expected, None)


def test_package_push_request_get_push_job_objects(package_push_request):
    expected = f"PackagePushRequestId = '{SF_ID}'"
    package_push_request.get_push_job_objs()
    package_push_request.push_api.get_push_job_objs.assert_called_once_with(
        expected, None
    )


def test_package_push_request_get_push_jobs_by_id(package_push_request):
    expected = f"PackagePushRequestId = '{SF_ID}'"
    package_push_request.get_push_jobs_by_id()
    package_push_request.push_api.get_push_jobs_by_id.assert_called_once_with(
        expected, None
    )


def test_format_where(package_subscriber):
    assert package_subscriber.format_where("foo") == "foo = 'bar'"
    assert (
        package_subscriber.format_where("foo", "foobar") == "foo = 'bar' AND (foobar)"
    )


def test_package_subscriber_get_push_jobs(package_subscriber):
    expected = f"SubscriberOrganizationKey = '{ORG_KEY}'"
    package_subscriber.get_push_jobs()
    package_subscriber.push_api.get_push_jobs.assert_called_once_with(expected, None)


def test_package_subscriber_get_push_job_objects(package_subscriber):
    expected = f"SubscriberOrganizationKey = '{ORG_KEY}'"
    package_subscriber.get_push_job_objs()
    package_subscriber.push_api.get_push_job_objs.assert_called_once_with(
        expected, None
    )


def test_package_subscriber_get_push_jobs_by_id(package_subscriber):
    expected = f"SubscriberOrganizationKey = '{ORG_KEY}'"
    package_subscriber.get_push_jobs_by_id()
    package_subscriber.push_api.get_push_jobs_by_id.assert_called_once_with(
        expected, None
    )
